# Copyright 2022 openStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from base64 import b64encode
import http
from http import client
from unittest import mock
from urllib import parse

from cryptography.hazmat.primitives.serialization import Encoding
import fixtures
from oslo_log import log
from oslo_serialization import jsonutils

from keystone.api.os_oauth2 import AccessTokenResource
from keystone.common import provider_api
from keystone.common import utils
from keystone import conf
from keystone import exception
from keystone.federation.utils import RuleProcessor
from keystone.tests import unit
from keystone.tests.unit import test_v3
from keystone.token.provider import Manager

PROVIDERS = provider_api.ProviderAPIs
LOG = log.getLogger(__name__)
CONF = conf.CONF


class FakeUserAppCredListCreateResource(mock.Mock):
    pass


class OAuth2AuthnMethodsTests(test_v3.OAuth2RestfulTestCase):
    ACCESS_TOKEN_URL = '/OS-OAUTH2/token'

    def setUp(self):
        super().setUp()
        self.config_fixture.config(
            group='oauth2',
            oauth2_authn_methods=['client_secret_basic', 'tls_client_auth'],
        )

    def _get_access_token(
        self, headers, data, expected_status, client_cert_content=None
    ):
        data = parse.urlencode(data).encode()
        kwargs = {
            'headers': headers,
            'noauth': True,
            'convert': False,
            'body': data,
            'expected_status': expected_status,
        }
        if client_cert_content:
            kwargs.update(
                {'environ': {'SSL_CLIENT_CERT': client_cert_content}}
            )
        resp = self.post(self.ACCESS_TOKEN_URL, **kwargs)
        return resp

    def _create_certificates(self):
        return unit.create_certificate(
            subject_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='tokyo',
                locality_name='musashino',
                organizational_unit_name='test',
            )
        )

    def _get_cert_content(self, cert):
        return cert.public_bytes(Encoding.PEM).decode('ascii')

    @mock.patch.object(AccessTokenResource, '_client_secret_basic')
    def test_secret_basic_header(self, mock_client_secret_basic):
        """client_secret_basic is used if a client sercret is found."""
        client_id = 'client_id'
        client_secret = 'client_secret'
        b64str = (
            b64encode(f'{client_id}:{client_secret}'.encode()).decode().strip()
        )
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': f'Basic {b64str}',
        }
        data = {'grant_type': 'client_credentials'}

        _ = self._get_access_token(
            headers=headers, data=data, expected_status=client.OK
        )
        mock_client_secret_basic.assert_called_once_with(
            client_id, client_secret
        )

    @mock.patch.object(AccessTokenResource, '_client_secret_basic')
    def test_secret_basic_form(self, mock_client_secret_basic):
        """client_secret_basic is used if a client sercret is found."""
        client_id = 'client_id'
        client_secret = 'client_secret'
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        data = {
            'grant_type': 'client_credentials',
            'client_id': client_id,
            'client_secret': client_secret,
        }

        _ = self._get_access_token(
            headers=headers, data=data, expected_status=client.OK
        )
        mock_client_secret_basic.assert_called_once_with(
            client_id, client_secret
        )

    @mock.patch.object(AccessTokenResource, '_client_secret_basic')
    def test_secret_basic_header_and_form(self, mock_client_secret_basic):
        """A header is used if secrets are found in a header and body."""
        client_id_h = 'client_id_h'
        client_secret_h = 'client_secret_h'
        client_id_d = 'client_id_d'
        client_secret_d = 'client_secret_d'
        b64str = (
            b64encode(f'{client_id_h}:{client_secret_h}'.encode())
            .decode()
            .strip()
        )
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': f'Basic {b64str}',
        }
        data = {
            'grant_type': 'client_credentials',
            'client_id': client_id_d,
            'client_secret': client_secret_d,
        }

        _ = self._get_access_token(
            headers=headers, data=data, expected_status=client.OK
        )
        mock_client_secret_basic.assert_called_once_with(
            client_id_h, client_secret_h
        )

    @mock.patch.object(AccessTokenResource, '_tls_client_auth')
    def test_client_cert(self, mock_tls_client_auth):
        """tls_client_auth is used if a certificate is found."""
        client_id = 'client_id'
        client_cert, _ = self._create_certificates()
        cert_content = self._get_cert_content(client_cert)
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        data = {'grant_type': 'client_credentials', 'client_id': client_id}
        _ = self._get_access_token(
            headers=headers,
            data=data,
            expected_status=client.OK,
            client_cert_content=cert_content,
        )
        mock_tls_client_auth.assert_called_once_with(client_id, cert_content)

    @mock.patch.object(AccessTokenResource, '_tls_client_auth')
    def test_secret_basic_and_client_cert(self, mock_tls_client_auth):
        """tls_client_auth is used if a certificate and secret are found."""
        client_id_s = 'client_id_s'
        client_secret = 'client_secret'
        client_id_c = 'client_id_c'
        client_cert, _ = self._create_certificates()
        cert_content = self._get_cert_content(client_cert)
        b64str = (
            b64encode(f'{client_id_s}:{client_secret}'.encode())
            .decode()
            .strip()
        )
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': f'Basic {b64str}',
        }
        data = {'grant_type': 'client_credentials', 'client_id': client_id_c}

        _ = self._get_access_token(
            headers=headers,
            data=data,
            expected_status=client.OK,
            client_cert_content=cert_content,
        )
        mock_tls_client_auth.assert_called_once_with(client_id_c, cert_content)


class OAuth2SecretBasicTests(test_v3.OAuth2RestfulTestCase):
    APP_CRED_CREATE_URL = '/users/%(user_id)s/application_credentials'
    APP_CRED_LIST_URL = '/users/%(user_id)s/application_credentials'
    APP_CRED_DELETE_URL = (
        '/users/%(user_id)s/application_credentials/%(app_cred_id)s'
    )
    APP_CRED_SHOW_URL = (
        '/users/%(user_id)s/application_credentials/%(app_cred_id)s'
    )
    ACCESS_TOKEN_URL = '/OS-OAUTH2/token'

    def setUp(self):
        super().setUp()
        log.set_defaults(
            logging_context_format_string='%(asctime)s.%(msecs)03d %('
            'color)s%(levelname)s %(name)s [^[['
            '01;36m%(request_id)s ^[[00;36m%('
            'project_name)s %(user_name)s%('
            'color)s] ^[[01;35m%(instance)s%('
            'color)s%(message)s^[[00m',
            default_log_levels=log.DEBUG,
        )
        CONF.log_opt_values(LOG, log.DEBUG)
        LOG.debug(f'is_debug_enabled: {log.is_debug_enabled(CONF)}')
        LOG.debug(f'get_default_log_levels: {log.get_default_log_levels()}')
        self.config_fixture.config(
            group='oauth2', oauth2_authn_methods=['client_secret_basic']
        )

    def _assert_error_resp(self, error_resp, error_msg, error_description):
        resp_keys = ('error', 'error_description')
        for key in resp_keys:
            self.assertIsNotNone(error_resp.get(key, None))
        self.assertEqual(error_msg, error_resp.get('error'))
        self.assertEqual(
            error_description, error_resp.get('error_description')
        )

    def _create_app_cred(self, user_id, app_cred_name):
        resp = self.post(
            self.APP_CRED_CREATE_URL % {'user_id': user_id},
            body={'application_credential': {'name': app_cred_name}},
        )
        LOG.debug(f'resp: {resp}')
        app_ref = resp.result['application_credential']
        return app_ref

    def _delete_app_cred(self, user_id, app_cred_id):
        resp = self.delete(
            self.APP_CRED_CREATE_URL
            % {'user_id': user_id, 'app_cred_id': app_cred_id}
        )
        LOG.debug(f'resp: {resp}')

    def _get_access_token(
        self, app_cred, b64str, headers, data, expected_status
    ):
        if b64str is None:
            client_id = app_cred.get('id')
            client_secret = app_cred.get('secret')
            b64str = (
                b64encode(f'{client_id}:{client_secret}'.encode())
                .decode()
                .strip()
            )
        if headers is None:
            headers = {
                'Content-Type': 'application/x-www-form-urlencoded',
                'Authorization': f'Basic {b64str}',
            }
        if data is None:
            data = {'grant_type': 'client_credentials'}
        data = parse.urlencode(data).encode()
        resp = self.post(
            self.ACCESS_TOKEN_URL,
            headers=headers,
            convert=False,
            body=data,
            expected_status=expected_status,
        )
        return resp

    def _get_access_token_method_not_allowed(self, app_cred, http_func):
        client_id = app_cred.get('id')
        client_secret = app_cred.get('secret')
        b64str = (
            b64encode(f'{client_id}:{client_secret}'.encode()).decode().strip()
        )
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': f'Basic {b64str}',
        }
        data = {'grant_type': 'client_credentials'}
        data = parse.urlencode(data).encode()
        resp = http_func(
            self.ACCESS_TOKEN_URL,
            headers=headers,
            convert=False,
            body=data,
            expected_status=client.METHOD_NOT_ALLOWED,
        )
        LOG.debug(f'response: {resp}')
        json_resp = jsonutils.loads(resp.body)
        return json_resp

    def test_get_access_token(self):
        """Test case when an access token can be successfully obtain."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        resp = self._get_access_token(
            app_cred,
            b64str=None,
            headers=None,
            data=None,
            expected_status=client.OK,
        )
        json_resp = jsonutils.loads(resp.body)
        self.assertIn('access_token', json_resp)
        self.assertEqual('Bearer', json_resp['token_type'])
        self.assertEqual(3600, json_resp['expires_in'])

    def test_get_access_token_form(self):
        """Test case when there is no client authorization."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        data = {
            'grant_type': 'client_credentials',
            'client_id': app_cred.get('id'),
            'client_secret': app_cred.get('secret'),
        }
        resp = self._get_access_token(
            app_cred,
            b64str=None,
            headers=headers,
            data=data,
            expected_status=client.OK,
        )
        json_resp = jsonutils.loads(resp.body)
        self.assertIn('access_token', json_resp)
        self.assertEqual('Bearer', json_resp['token_type'])
        self.assertEqual(3600, json_resp['expires_in'])

    def test_get_access_token_auth_type_is_not_basic(self):
        """Test case when auth_type is not basic."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        client_id = app_cred.get('id')

        base = (
            'username="{}", realm="{}", nonce="{}", uri="{}", '
            'response="{}"'.format(
                client_id, 'realm', 'nonce', 'path', 'responding'
            )
        )

        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': f'Digest {base}',
        }
        error = 'invalid_client'
        error_description = 'Client authentication failed.'
        resp = self._get_access_token(
            app_cred,
            b64str=None,
            headers=headers,
            data=None,
            expected_status=client.UNAUTHORIZED,
        )
        self.assertNotEmpty(resp.headers.get("WWW-Authenticate"))
        self.assertEqual(
            'Keystone uri="http://localhost/v3"',
            resp.headers.get("WWW-Authenticate"),
        )
        json_resp = jsonutils.loads(resp.body)
        LOG.debug(f'error: {json_resp.get("error")}')
        LOG.debug(f'error_description: {json_resp.get("error_description")}')
        self.assertEqual(error, json_resp.get('error'))
        self.assertEqual(error_description, json_resp.get('error_description'))

    def test_get_access_token_without_client_id(self):
        """Test case when there is no client_id."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        client_secret = app_cred.get('secret')
        b64str = b64encode(f':{client_secret}'.encode()).decode().strip()
        error = 'invalid_client'
        error_description = 'Client authentication failed.'
        resp = self._get_access_token(
            app_cred,
            b64str=b64str,
            headers=None,
            data=None,
            expected_status=client.UNAUTHORIZED,
        )
        self.assertNotEmpty(resp.headers.get("WWW-Authenticate"))
        self.assertEqual(
            'Keystone uri="http://localhost/v3"',
            resp.headers.get("WWW-Authenticate"),
        )
        json_resp = jsonutils.loads(resp.body)
        LOG.debug(f'error: {json_resp.get("error")}')
        LOG.debug(f'error_description: {json_resp.get("error_description")}')
        self.assertEqual(error, json_resp.get('error'))
        self.assertEqual(error_description, json_resp.get('error_description'))

    def test_get_access_token_without_client_secret(self):
        """Test case when there is no client_secret."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        client_id = app_cred.get('id')
        b64str = b64encode(f'{client_id}:'.encode()).decode().strip()
        error = 'invalid_client'
        error_description = 'Client authentication failed.'
        resp = self._get_access_token(
            app_cred,
            b64str=b64str,
            headers=None,
            data=None,
            expected_status=client.UNAUTHORIZED,
        )
        self.assertNotEmpty(resp.headers.get("WWW-Authenticate"))
        self.assertEqual(
            'Keystone uri="http://localhost/v3"',
            resp.headers.get("WWW-Authenticate"),
        )
        json_resp = jsonutils.loads(resp.body)
        LOG.debug(f'error: {json_resp.get("error")}')
        LOG.debug(f'error_description: {json_resp.get("error_description")}')
        self.assertEqual(error, json_resp.get('error'))
        self.assertEqual(error_description, json_resp.get('error_description'))

    def test_get_access_token_without_grant_type(self):
        """Test case when there is no grant_type."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        data = {}
        error = 'invalid_request'
        error_description = 'The parameter grant_type is required.'
        resp = self._get_access_token(
            app_cred,
            b64str=None,
            headers=None,
            data=data,
            expected_status=client.BAD_REQUEST,
        )
        json_resp = jsonutils.loads(resp.body)
        LOG.debug(f'error: {json_resp.get("error")}')
        LOG.debug(f'error_description: {json_resp.get("error_description")}')
        self.assertEqual(error, json_resp.get('error'))
        self.assertEqual(error_description, json_resp.get('error_description'))

    def test_get_access_token_blank_grant_type(self):
        """Test case when grant_type is blank."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        data = {'grant_type': ''}
        error = 'unsupported_grant_type'
        error_description = (
            f'The parameter grant_type {data["grant_type"]} is not supported.'
        )
        resp = self._get_access_token(
            app_cred,
            b64str=None,
            headers=None,
            data=data,
            expected_status=client.BAD_REQUEST,
        )
        json_resp = jsonutils.loads(resp.body)
        LOG.debug(f'error: {json_resp.get("error")}')
        LOG.debug(f'error_description: {json_resp.get("error_description")}')
        self.assertEqual(error, json_resp.get('error'))
        self.assertEqual(error_description, json_resp.get('error_description'))

    def test_get_access_token_grant_type_is_not_client_credentials(self):
        """Test case when grant_type is not client_credentials."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        data = {'grant_type': 'not_client_credentials'}
        error = 'unsupported_grant_type'
        error_description = (
            f'The parameter grant_type {data["grant_type"]} is not supported.'
        )
        resp = self._get_access_token(
            app_cred,
            b64str=None,
            headers=None,
            data=data,
            expected_status=client.BAD_REQUEST,
        )
        json_resp = jsonutils.loads(resp.body)
        LOG.debug(f'error: {json_resp.get("error")}')
        LOG.debug(f'error_description: {json_resp.get("error_description")}')
        self.assertEqual(error, json_resp.get('error'))
        self.assertEqual(error_description, json_resp.get('error_description'))

    def test_get_access_token_failed_401(self):
        """Test case when client authentication failed."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        error = 'invalid_client'

        client_id = app_cred.get('id')
        client_secret = app_cred.get('secret')
        b64str = (
            b64encode(f'{client_id}:{client_secret}'.encode()).decode().strip()
        )
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': f'Basic {b64str}',
        }
        data = {'grant_type': 'client_credentials'}
        data = parse.urlencode(data).encode()
        with mock.patch(
            'keystone.api._shared.authentication.authenticate_for_token'
        ) as co_mock:
            co_mock.side_effect = exception.Unauthorized(
                'client is unauthorized'
            )
            resp = self.post(
                self.ACCESS_TOKEN_URL,
                headers=headers,
                convert=False,
                body=data,
                noauth=True,
                expected_status=client.UNAUTHORIZED,
            )
            self.assertNotEmpty(resp.headers.get("WWW-Authenticate"))
            self.assertEqual(
                'Keystone uri="http://localhost/v3"',
                resp.headers.get("WWW-Authenticate"),
            )
        LOG.debug(f'response: {resp}')
        json_resp = jsonutils.loads(resp.body)
        self.assertEqual(error, json_resp.get('error'))
        LOG.debug(f'error: {json_resp.get("error")}')

    def test_get_access_token_failed_400(self):
        """Test case when the called API is incorrect."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        error = 'invalid_request'
        client_id = app_cred.get('id')
        client_secret = app_cred.get('secret')
        b64str = (
            b64encode(f'{client_id}:{client_secret}'.encode()).decode().strip()
        )
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': f'Basic {b64str}',
        }
        data = {'grant_type': 'client_credentials'}
        data = parse.urlencode(data).encode()
        with mock.patch(
            'keystone.api._shared.authentication.authenticate_for_token'
        ) as co_mock:
            co_mock.side_effect = exception.ValidationError(
                'Auth method is invalid'
            )
            resp = self.post(
                self.ACCESS_TOKEN_URL,
                headers=headers,
                convert=False,
                body=data,
                noauth=True,
                expected_status=client.BAD_REQUEST,
            )
            LOG.debug(f'response: {resp}')
            json_resp = jsonutils.loads(resp.body)
            self.assertEqual(error, json_resp.get('error'))
            LOG.debug(f'error: {json_resp.get("error")}')

    def test_get_access_token_failed_500_other(self):
        """Test case when unexpected error."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        error = 'other_error'
        client_id = app_cred.get('id')
        client_secret = app_cred.get('secret')
        b64str = (
            b64encode(f'{client_id}:{client_secret}'.encode()).decode().strip()
        )
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': f'Basic {b64str}',
        }
        data = {'grant_type': 'client_credentials'}
        data = parse.urlencode(data).encode()
        with mock.patch(
            'keystone.api._shared.authentication.authenticate_for_token'
        ) as co_mock:
            co_mock.side_effect = exception.UnexpectedError(
                'unexpected error.'
            )
            resp = self.post(
                self.ACCESS_TOKEN_URL,
                headers=headers,
                convert=False,
                body=data,
                noauth=True,
                expected_status=client.INTERNAL_SERVER_ERROR,
            )

        LOG.debug(f'response: {resp}')
        json_resp = jsonutils.loads(resp.body)
        self.assertEqual(error, json_resp.get('error'))

    def test_get_access_token_failed_500(self):
        """Test case when internal server error."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        error = 'other_error'
        client_id = app_cred.get('id')
        client_secret = app_cred.get('secret')
        b64str = (
            b64encode(f'{client_id}:{client_secret}'.encode()).decode().strip()
        )
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': f'Basic {b64str}',
        }
        data = {'grant_type': 'client_credentials'}
        data = parse.urlencode(data).encode()
        with mock.patch(
            'keystone.api._shared.authentication.authenticate_for_token'
        ) as co_mock:
            co_mock.side_effect = Exception('Internal server is invalid')
            resp = self.post(
                self.ACCESS_TOKEN_URL,
                headers=headers,
                convert=False,
                body=data,
                noauth=True,
                expected_status=client.INTERNAL_SERVER_ERROR,
            )

        LOG.debug(f'response: {resp}')
        json_resp = jsonutils.loads(resp.body)
        self.assertEqual(error, json_resp.get('error'))

    def test_get_access_token_method_get_not_allowed(self):
        """Test case when the request is get method that is not allowed."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        json_resp = self._get_access_token_method_not_allowed(
            app_cred, self.get
        )
        self.assertEqual('other_error', json_resp.get('error'))
        self.assertEqual(
            'The method is not allowed for the requested URL.',
            json_resp.get('error_description'),
        )

    def test_get_access_token_method_patch_not_allowed(self):
        """Test case when the request is patch method that is not allowed."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        json_resp = self._get_access_token_method_not_allowed(
            app_cred, self.patch
        )
        self.assertEqual('other_error', json_resp.get('error'))
        self.assertEqual(
            'The method is not allowed for the requested URL.',
            json_resp.get('error_description'),
        )

    def test_get_access_token_method_put_not_allowed(self):
        """Test case when the request is put method that is not allowed."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        json_resp = self._get_access_token_method_not_allowed(
            app_cred, self.put
        )
        self.assertEqual('other_error', json_resp.get('error'))
        self.assertEqual(
            'The method is not allowed for the requested URL.',
            json_resp.get('error_description'),
        )

    def test_get_access_token_method_delete_not_allowed(self):
        """Test case when the request is delete method that is not allowed."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        json_resp = self._get_access_token_method_not_allowed(
            app_cred, self.delete
        )
        self.assertEqual('other_error', json_resp.get('error'))
        self.assertEqual(
            'The method is not allowed for the requested URL.',
            json_resp.get('error_description'),
        )

    def test_get_access_token_method_head_not_allowed(self):
        """Test case when the request is head method that is not allowed."""
        client_name = 'client_name_test'
        app_cred = self._create_app_cred(self.user_id, client_name)
        client_id = app_cred.get('id')
        client_secret = app_cred.get('secret')
        b64str = (
            b64encode(f'{client_id}:{client_secret}'.encode()).decode().strip()
        )
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': f'Basic {b64str}',
        }
        self.head(
            self.ACCESS_TOKEN_URL,
            headers=headers,
            convert=False,
            expected_status=client.METHOD_NOT_ALLOWED,
        )


class OAuth2CertificateTests(test_v3.OAuth2RestfulTestCase):
    ACCESS_TOKEN_URL = '/OS-OAUTH2/token'

    def setUp(self):
        super().setUp()
        self.log_fix = self.useFixture(fixtures.FakeLogger(level=log.DEBUG))
        self.config_fixture.config(
            group='oauth2', oauth2_authn_methods=['tls_client_auth']
        )
        self.config_fixture.config(
            group='oauth2', oauth2_cert_dn_mapping_id='oauth2_mapping'
        )
        (self.oauth2_user, self.oauth2_user_domain, _) = (
            self._create_project_user()
        )
        *_, self.client_cert, self.client_key = self._create_certificates(
            client_dn=unit.create_dn(
                user_id=self.oauth2_user.get('id'),
                common_name=self.oauth2_user.get('name'),
                email_address=self.oauth2_user.get('email'),
                domain_component=self.oauth2_user_domain.get('id'),
                organization_name=self.oauth2_user_domain.get('name'),
            )
        )

    def _create_project_user(self, no_roles=False):
        new_domain_ref = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(
            new_domain_ref['id'], new_domain_ref
        )
        new_project_ref = unit.new_project_ref(domain_id=self.domain_id)
        PROVIDERS.resource_api.create_project(
            new_project_ref['id'], new_project_ref
        )
        new_user = unit.create_user(
            PROVIDERS.identity_api,
            domain_id=new_domain_ref['id'],
            project_id=new_project_ref['id'],
        )
        if not no_roles:
            PROVIDERS.assignment_api.create_grant(
                self.role['id'],
                user_id=new_user['id'],
                project_id=new_project_ref['id'],
            )
        return new_user, new_domain_ref, new_project_ref

    def _create_certificates(
        self, root_dn=None, server_dn=None, client_dn=None
    ):
        root_subj = unit.create_dn(
            country_name='jp',
            state_or_province_name='kanagawa',
            locality_name='kawasaki',
            organization_name='fujitsu',
            organizational_unit_name='test',
            common_name='root',
        )
        if root_dn:
            root_subj = unit.update_dn(root_subj, root_dn)

        root_cert, root_key = unit.create_certificate(root_subj)
        keystone_subj = unit.create_dn(
            country_name='jp',
            state_or_province_name='kanagawa',
            locality_name='kawasaki',
            organization_name='fujitsu',
            organizational_unit_name='test',
            common_name='keystone.local',
        )
        if server_dn:
            keystone_subj = unit.update_dn(keystone_subj, server_dn)

        ks_cert, ks_key = unit.create_certificate(
            keystone_subj, ca=root_cert, ca_key=root_key
        )
        client_subj = unit.create_dn(
            country_name='jp',
            state_or_province_name='kanagawa',
            locality_name='kawasaki',
            organizational_unit_name='test',
        )
        if client_dn:
            client_subj = unit.update_dn(client_subj, client_dn)

        client_cert, client_key = unit.create_certificate(
            client_subj, ca=root_cert, ca_key=root_key
        )
        return root_cert, root_key, ks_cert, ks_key, client_cert, client_key

    def _create_mapping(self, id='oauth2_mapping', dn_rules=None):
        rules = []
        if not dn_rules:
            dn_rules = [
                {
                    'user.name': 'SSL_CLIENT_SUBJECT_DN_CN',
                    'user.id': 'SSL_CLIENT_SUBJECT_DN_UID',
                    'user.email': 'SSL_CLIENT_SUBJECT_DN_EMAILADDRESS',
                    'user.domain.id': 'SSL_CLIENT_SUBJECT_DN_DC',
                    'user.domain.name': 'SSL_CLIENT_SUBJECT_DN_O',
                    'SSL_CLIENT_ISSUER_DN_CN': ['root'],
                }
            ]
        for info in dn_rules:
            index = 0
            local_user = {}
            remote = []
            for k in info:
                if k == 'user.name':
                    local_user['name'] = f'{{{index}}}'
                    remote.append({'type': info.get(k)})
                    index += 1
                elif k == 'user.id':
                    local_user['id'] = f'{{{index}}}'
                    remote.append({'type': info.get(k)})
                    index += 1
                elif k == 'user.email':
                    local_user['email'] = f'{{{index}}}'
                    remote.append({'type': info.get(k)})
                    index += 1
                elif k == 'user.domain.name' or k == 'user.domain.id':
                    if not local_user.get('domain'):
                        local_user['domain'] = {}
                    if k == 'user.domain.name':
                        local_user['domain']['name'] = f'{{{index}}}'
                        remote.append({'type': info.get(k)})
                        index += 1
                    else:
                        local_user['domain']['id'] = f'{{{index}}}'
                        remote.append({'type': info.get(k)})
                        index += 1
                else:
                    remote.append({'type': k, 'any_one_of': info.get(k)})
            rule = {'local': [{'user': local_user}], 'remote': remote}
            rules.append(rule)

        mapping = {'id': id, 'rules': rules}

        PROVIDERS.federation_api.create_mapping(mapping['id'], mapping)

    def _get_access_token(
        self,
        client_id=None,
        client_cert_content=None,
        expected_status=http.client.OK,
    ):
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        data = {'grant_type': 'client_credentials'}
        if client_id:
            data.update({'client_id': client_id})
        data = parse.urlencode(data).encode()
        kwargs = {
            'headers': headers,
            'noauth': True,
            'convert': False,
            'body': data,
            'expected_status': expected_status,
        }
        if client_cert_content:
            kwargs.update(
                {'environ': {'SSL_CLIENT_CERT': client_cert_content}}
            )
        resp = self.post(self.ACCESS_TOKEN_URL, **kwargs)
        return resp

    def _get_cert_content(self, cert):
        return cert.public_bytes(Encoding.PEM).decode('ascii')

    def assertUnauthorizedResp(self, resp):
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertEqual('invalid_client', json_resp['error'])
        self.assertEqual(
            'Client authentication failed.', json_resp['error_description']
        )

    def test_get_access_token_project_scope(self):
        """Test case when an access token can be successfully obtain."""
        self._create_mapping()
        user, user_domain, user_project = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'), client_cert_content=cert_content
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertIn('access_token', json_resp)
        self.assertEqual('Bearer', json_resp['token_type'])
        self.assertEqual(3600, json_resp['expires_in'])

        verify_resp = self.get(
            '/auth/tokens',
            headers={
                'X-Subject-Token': json_resp['access_token'],
                'X-Auth-Token': json_resp['access_token'],
            },
        )
        self.assertIn('token', verify_resp.result)
        self.assertIn('oauth2_credential', verify_resp.result['token'])
        self.assertIn('roles', verify_resp.result['token'])
        self.assertIn('project', verify_resp.result['token'])
        self.assertIn('catalog', verify_resp.result['token'])
        self.assertEqual(
            user_project.get('id'),
            verify_resp.result['token']['project']['id'],
        )
        check_oauth2 = verify_resp.result['token']['oauth2_credential']
        self.assertEqual(
            utils.get_certificate_thumbprint(cert_content),
            check_oauth2['x5t#S256'],
        )

    def test_get_access_token_mapping_config(self):
        """Test case when an access token can be successfully obtain."""
        self.config_fixture.config(
            group='oauth2', oauth2_cert_dn_mapping_id='oauth2_custom'
        )
        self._create_mapping(
            id='oauth2_custom',
            dn_rules=[
                {
                    'user.name': 'SSL_CLIENT_SUBJECT_DN_CN',
                    'user.domain.name': 'SSL_CLIENT_SUBJECT_DN_DC',
                    'SSL_CLIENT_ISSUER_DN_CN': ['root'],
                }
            ],
        )

        user, user_domain, user_project = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id='test_UID',
                common_name=user.get('name'),
                domain_component=user_domain.get('name'),
                organization_name='test_O',
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'), client_cert_content=cert_content
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertIn('access_token', json_resp)
        self.assertEqual('Bearer', json_resp['token_type'])
        self.assertEqual(3600, json_resp['expires_in'])

        verify_resp = self.get(
            '/auth/tokens',
            headers={
                'X-Subject-Token': json_resp['access_token'],
                'X-Auth-Token': json_resp['access_token'],
            },
        )
        self.assertIn('token', verify_resp.result)
        self.assertIn('oauth2_credential', verify_resp.result['token'])
        self.assertIn('roles', verify_resp.result['token'])
        self.assertIn('project', verify_resp.result['token'])
        self.assertIn('catalog', verify_resp.result['token'])
        self.assertEqual(
            user_project.get('id'),
            verify_resp.result['token']['project']['id'],
        )
        check_oauth2 = verify_resp.result['token']['oauth2_credential']
        self.assertEqual(
            utils.get_certificate_thumbprint(cert_content),
            check_oauth2['x5t#S256'],
        )

        self.config_fixture.config(
            group='oauth2', oauth2_cert_dn_mapping_id='oauth2_mapping'
        )

    def test_get_access_token_mapping_multi_ca(self):
        """Test case when an access token can be successfully obtain."""
        self.config_fixture.config(
            group='oauth2', oauth2_cert_dn_mapping_id='oauth2_custom'
        )
        self._create_mapping(
            id='oauth2_custom',
            dn_rules=[
                {
                    'user.name': 'SSL_CLIENT_SUBJECT_DN_CN',
                    'user.id': 'SSL_CLIENT_SUBJECT_DN_UID',
                    'user.email': 'SSL_CLIENT_SUBJECT_DN_EMAILADDRESS',
                    'user.domain.id': 'SSL_CLIENT_SUBJECT_DN_DC',
                    'user.domain.name': 'SSL_CLIENT_SUBJECT_DN_O',
                    'SSL_CLIENT_ISSUER_DN_CN': ['rootA', 'rootB'],
                },
                {
                    'user.name': 'SSL_CLIENT_SUBJECT_DN_CN',
                    'user.domain.name': 'SSL_CLIENT_SUBJECT_DN_DC',
                    'SSL_CLIENT_ISSUER_DN_CN': ['rootC'],
                },
            ],
        )

        # CA rootA OK
        user, user_domain, user_project = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            root_dn=unit.create_dn(common_name='rootA'),
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            ),
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'), client_cert_content=cert_content
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertIn('access_token', json_resp)
        self.assertEqual('Bearer', json_resp['token_type'])
        self.assertEqual(3600, json_resp['expires_in'])

        verify_resp = self.get(
            '/auth/tokens',
            headers={
                'X-Subject-Token': json_resp['access_token'],
                'X-Auth-Token': json_resp['access_token'],
            },
        )
        self.assertIn('token', verify_resp.result)
        self.assertIn('oauth2_credential', verify_resp.result['token'])
        self.assertIn('roles', verify_resp.result['token'])
        self.assertIn('project', verify_resp.result['token'])
        self.assertIn('catalog', verify_resp.result['token'])
        self.assertEqual(
            user_project.get('id'),
            verify_resp.result['token']['project']['id'],
        )
        check_oauth2 = verify_resp.result['token']['oauth2_credential']
        self.assertEqual(
            utils.get_certificate_thumbprint(cert_content),
            check_oauth2['x5t#S256'],
        )

        # CA rootB OK
        user, user_domain, user_project = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            root_dn=unit.create_dn(common_name='rootB'),
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            ),
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'), client_cert_content=cert_content
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertIn('access_token', json_resp)
        self.assertEqual('Bearer', json_resp['token_type'])
        self.assertEqual(3600, json_resp['expires_in'])

        verify_resp = self.get(
            '/auth/tokens',
            headers={
                'X-Subject-Token': json_resp['access_token'],
                'X-Auth-Token': json_resp['access_token'],
            },
        )
        self.assertIn('token', verify_resp.result)
        self.assertIn('oauth2_credential', verify_resp.result['token'])
        self.assertIn('roles', verify_resp.result['token'])
        self.assertIn('project', verify_resp.result['token'])
        self.assertIn('catalog', verify_resp.result['token'])
        self.assertEqual(
            user_project.get('id'),
            verify_resp.result['token']['project']['id'],
        )
        check_oauth2 = verify_resp.result['token']['oauth2_credential']
        self.assertEqual(
            utils.get_certificate_thumbprint(cert_content),
            check_oauth2['x5t#S256'],
        )

        # CA rootC OK
        user, user_domain, user_project = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            root_dn=unit.create_dn(common_name='rootC'),
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id='test_UID',
                common_name=user.get('name'),
                domain_component=user_domain.get('name'),
                organization_name='test_O',
            ),
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'), client_cert_content=cert_content
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertIn('access_token', json_resp)
        self.assertEqual('Bearer', json_resp['token_type'])
        self.assertEqual(3600, json_resp['expires_in'])

        verify_resp = self.get(
            '/auth/tokens',
            headers={
                'X-Subject-Token': json_resp['access_token'],
                'X-Auth-Token': json_resp['access_token'],
            },
        )
        self.assertIn('token', verify_resp.result)
        self.assertIn('oauth2_credential', verify_resp.result['token'])
        self.assertIn('roles', verify_resp.result['token'])
        self.assertIn('project', verify_resp.result['token'])
        self.assertIn('catalog', verify_resp.result['token'])
        self.assertEqual(
            user_project.get('id'),
            verify_resp.result['token']['project']['id'],
        )
        check_oauth2 = verify_resp.result['token']['oauth2_credential']
        self.assertEqual(
            utils.get_certificate_thumbprint(cert_content),
            check_oauth2['x5t#S256'],
        )

        # CA not found NG
        user, user_domain, user_project = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            root_dn=unit.create_dn(common_name='root_other'),
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            ),
        )
        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: mapping rule process failed.',
            self.log_fix.output,
        )

        self.config_fixture.config(
            group='oauth2', oauth2_cert_dn_mapping_id='oauth2_mapping'
        )

    def test_get_access_token_no_default_mapping(self):
        user, user_domain, _ = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: '
            'mapping id {} is not found. '.format('oauth2_mapping'),
            self.log_fix.output,
        )

    def test_get_access_token_no_custom_mapping(self):
        self.config_fixture.config(
            group='oauth2', oauth2_cert_dn_mapping_id='oauth2_custom'
        )
        self._create_mapping()
        user, user_domain, _ = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: '
            'mapping id {} is not found. '.format('oauth2_custom'),
            self.log_fix.output,
        )
        self.config_fixture.config(
            group='oauth2', oauth2_cert_dn_mapping_id='oauth2_mapping'
        )

    def test_get_access_token_ignore_userid(self):
        """Test case when an access token can be successfully obtain."""
        self._create_mapping(
            dn_rules=[
                {
                    'user.name': 'SSL_CLIENT_SUBJECT_DN_CN',
                    'user.email': 'SSL_CLIENT_SUBJECT_DN_EMAILADDRESS',
                    'user.domain.id': 'SSL_CLIENT_SUBJECT_DN_DC',
                    'user.domain.name': 'SSL_CLIENT_SUBJECT_DN_O',
                    'SSL_CLIENT_ISSUER_DN_CN': ['root'],
                }
            ]
        )

        user, user_domain, user_project = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id') + "_diff",
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'), client_cert_content=cert_content
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertIn('access_token', json_resp)
        self.assertEqual('Bearer', json_resp['token_type'])
        self.assertEqual(3600, json_resp['expires_in'])

        verify_resp = self.get(
            '/auth/tokens',
            headers={
                'X-Subject-Token': json_resp['access_token'],
                'X-Auth-Token': json_resp['access_token'],
            },
        )
        self.assertIn('token', verify_resp.result)
        self.assertIn('oauth2_credential', verify_resp.result['token'])
        self.assertIn('roles', verify_resp.result['token'])
        self.assertIn('project', verify_resp.result['token'])
        self.assertIn('catalog', verify_resp.result['token'])
        self.assertEqual(
            user_project.get('id'),
            verify_resp.result['token']['project']['id'],
        )
        check_oauth2 = verify_resp.result['token']['oauth2_credential']
        self.assertEqual(
            utils.get_certificate_thumbprint(cert_content),
            check_oauth2['x5t#S256'],
        )

    def test_get_access_token_ignore_username(self):
        """Test case when an access token can be successfully obtain."""
        self._create_mapping(
            dn_rules=[
                {
                    'user.id': 'SSL_CLIENT_SUBJECT_DN_UID',
                    'user.email': 'SSL_CLIENT_SUBJECT_DN_EMAILADDRESS',
                    'user.domain.id': 'SSL_CLIENT_SUBJECT_DN_DC',
                    'user.domain.name': 'SSL_CLIENT_SUBJECT_DN_O',
                    'SSL_CLIENT_ISSUER_DN_CN': ['root'],
                }
            ]
        )
        user, user_domain, user_project = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'), client_cert_content=cert_content
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertIn('access_token', json_resp)
        self.assertEqual('Bearer', json_resp['token_type'])
        self.assertEqual(3600, json_resp['expires_in'])

        verify_resp = self.get(
            '/auth/tokens',
            headers={
                'X-Subject-Token': json_resp['access_token'],
                'X-Auth-Token': json_resp['access_token'],
            },
        )
        self.assertIn('token', verify_resp.result)
        self.assertIn('oauth2_credential', verify_resp.result['token'])
        self.assertIn('roles', verify_resp.result['token'])
        self.assertIn('project', verify_resp.result['token'])
        self.assertIn('catalog', verify_resp.result['token'])
        self.assertEqual(
            user_project.get('id'),
            verify_resp.result['token']['project']['id'],
        )
        check_oauth2 = verify_resp.result['token']['oauth2_credential']
        self.assertEqual(
            utils.get_certificate_thumbprint(cert_content),
            check_oauth2['x5t#S256'],
        )

    def test_get_access_token_ignore_email(self):
        """Test case when an access token can be successfully obtain."""
        self._create_mapping(
            dn_rules=[
                {
                    'user.name': 'SSL_CLIENT_SUBJECT_DN_CN',
                    'user.id': 'SSL_CLIENT_SUBJECT_DN_UID',
                    'user.domain.id': 'SSL_CLIENT_SUBJECT_DN_DC',
                    'user.domain.name': 'SSL_CLIENT_SUBJECT_DN_O',
                    'SSL_CLIENT_ISSUER_DN_CN': ['root'],
                }
            ]
        )
        user, user_domain, user_project = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'), client_cert_content=cert_content
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertIn('access_token', json_resp)
        self.assertEqual('Bearer', json_resp['token_type'])
        self.assertEqual(3600, json_resp['expires_in'])

        verify_resp = self.get(
            '/auth/tokens',
            headers={
                'X-Subject-Token': json_resp['access_token'],
                'X-Auth-Token': json_resp['access_token'],
            },
        )
        self.assertIn('token', verify_resp.result)
        self.assertIn('oauth2_credential', verify_resp.result['token'])
        self.assertIn('roles', verify_resp.result['token'])
        self.assertIn('project', verify_resp.result['token'])
        self.assertIn('catalog', verify_resp.result['token'])
        self.assertEqual(
            user_project.get('id'),
            verify_resp.result['token']['project']['id'],
        )
        check_oauth2 = verify_resp.result['token']['oauth2_credential']
        self.assertEqual(
            utils.get_certificate_thumbprint(cert_content),
            check_oauth2['x5t#S256'],
        )

    def test_get_access_token_ignore_domain_id(self):
        """Test case when an access token can be successfully obtain."""
        self._create_mapping(
            dn_rules=[
                {
                    'user.name': 'SSL_CLIENT_SUBJECT_DN_CN',
                    'user.id': 'SSL_CLIENT_SUBJECT_DN_UID',
                    'user.email': 'SSL_CLIENT_SUBJECT_DN_EMAILADDRESS',
                    'user.domain.name': 'SSL_CLIENT_SUBJECT_DN_O',
                    'SSL_CLIENT_ISSUER_DN_CN': ['root'],
                }
            ]
        )
        user, user_domain, user_project = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id') + "_diff",
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'), client_cert_content=cert_content
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertIn('access_token', json_resp)
        self.assertEqual('Bearer', json_resp['token_type'])
        self.assertEqual(3600, json_resp['expires_in'])

        verify_resp = self.get(
            '/auth/tokens',
            headers={
                'X-Subject-Token': json_resp['access_token'],
                'X-Auth-Token': json_resp['access_token'],
            },
        )
        self.assertIn('token', verify_resp.result)
        self.assertIn('oauth2_credential', verify_resp.result['token'])
        self.assertIn('roles', verify_resp.result['token'])
        self.assertIn('project', verify_resp.result['token'])
        self.assertIn('catalog', verify_resp.result['token'])
        self.assertEqual(
            user_project.get('id'),
            verify_resp.result['token']['project']['id'],
        )
        check_oauth2 = verify_resp.result['token']['oauth2_credential']
        self.assertEqual(
            utils.get_certificate_thumbprint(cert_content),
            check_oauth2['x5t#S256'],
        )

    def test_get_access_token_ignore_domain_name(self):
        """Test case when an access token can be successfully obtain."""
        self._create_mapping(
            dn_rules=[
                {
                    'user.name': 'SSL_CLIENT_SUBJECT_DN_CN',
                    'user.id': 'SSL_CLIENT_SUBJECT_DN_UID',
                    'user.email': 'SSL_CLIENT_SUBJECT_DN_EMAILADDRESS',
                    'user.domain.id': 'SSL_CLIENT_SUBJECT_DN_DC',
                    'SSL_CLIENT_ISSUER_DN_CN': ['root'],
                }
            ]
        )
        user, user_domain, user_project = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'), client_cert_content=cert_content
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertIn('access_token', json_resp)
        self.assertEqual('Bearer', json_resp['token_type'])
        self.assertEqual(3600, json_resp['expires_in'])

        verify_resp = self.get(
            '/auth/tokens',
            headers={
                'X-Subject-Token': json_resp['access_token'],
                'X-Auth-Token': json_resp['access_token'],
            },
        )
        self.assertIn('token', verify_resp.result)
        self.assertIn('oauth2_credential', verify_resp.result['token'])
        self.assertIn('roles', verify_resp.result['token'])
        self.assertIn('project', verify_resp.result['token'])
        self.assertIn('catalog', verify_resp.result['token'])
        self.assertEqual(
            user_project.get('id'),
            verify_resp.result['token']['project']['id'],
        )
        check_oauth2 = verify_resp.result['token']['oauth2_credential']
        self.assertEqual(
            utils.get_certificate_thumbprint(cert_content),
            check_oauth2['x5t#S256'],
        )

    def test_get_access_token_ignore_all(self):
        """Test case when an access token can be successfully obtain."""
        self._create_mapping(dn_rules=[{'SSL_CLIENT_ISSUER_DN_CN': ['root']}])
        user, user_domain, user_project = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id') + "_diff",
                common_name=user.get('name') + "_diff",
                email_address=user.get('email') + "_diff",
                domain_component=user_domain.get('id') + "_diff",
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'), client_cert_content=cert_content
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertIn('access_token', json_resp)
        self.assertEqual('Bearer', json_resp['token_type'])
        self.assertEqual(3600, json_resp['expires_in'])

        verify_resp = self.get(
            '/auth/tokens',
            headers={
                'X-Subject-Token': json_resp['access_token'],
                'X-Auth-Token': json_resp['access_token'],
            },
        )
        self.assertIn('token', verify_resp.result)
        self.assertIn('oauth2_credential', verify_resp.result['token'])
        self.assertIn('roles', verify_resp.result['token'])
        self.assertIn('project', verify_resp.result['token'])
        self.assertIn('catalog', verify_resp.result['token'])
        self.assertEqual(
            user_project.get('id'),
            verify_resp.result['token']['project']['id'],
        )
        check_oauth2 = verify_resp.result['token']['oauth2_credential']
        self.assertEqual(
            utils.get_certificate_thumbprint(cert_content),
            check_oauth2['x5t#S256'],
        )

    def test_get_access_token_no_roles_project_scope(self):
        self._create_mapping()
        user, user_domain, _ = self._create_project_user(no_roles=True)
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        LOG.debug(resp)

    def test_get_access_token_no_default_project_id(self):
        self._create_mapping()
        user, user_domain, _ = self._create_project_user(no_roles=True)
        user['default_project_id'] = None
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        _ = self._get_access_token(
            client_id=user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )

    def test_get_access_token_without_client_id(self):
        self._create_mapping()
        cert_content = self._get_cert_content(self.client_cert)
        resp = self._get_access_token(
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: '
            'failed to get a client_id from the request.',
            self.log_fix.output,
        )

    def test_get_access_token_without_client_cert(self):
        self._create_mapping()
        resp = self._get_access_token(
            client_id=self.oauth2_user.get('id'),
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: '
            'failed to get client credentials from the request.',
            self.log_fix.output,
        )

    @mock.patch.object(utils, 'get_certificate_subject_dn')
    def test_get_access_token_failed_to_get_cert_subject_dn(
        self, mock_get_certificate_subject_dn
    ):
        self._create_mapping()
        mock_get_certificate_subject_dn.side_effect = (
            exception.ValidationError('Boom!')
        )
        cert_content = self._get_cert_content(self.client_cert)
        resp = self._get_access_token(
            client_id=self.oauth2_user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: '
            'failed to get the subject DN from the certificate.',
            self.log_fix.output,
        )

    @mock.patch.object(utils, 'get_certificate_issuer_dn')
    def test_get_access_token_failed_to_get_cert_issuer_dn(
        self, mock_get_certificate_issuer_dn
    ):
        self._create_mapping()
        mock_get_certificate_issuer_dn.side_effect = exception.ValidationError(
            'Boom!'
        )
        cert_content = self._get_cert_content(self.client_cert)
        resp = self._get_access_token(
            client_id=self.oauth2_user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: '
            'failed to get the issuer DN from the certificate.',
            self.log_fix.output,
        )

    def test_get_access_token_user_not_exist(self):
        self._create_mapping()
        cert_content = self._get_cert_content(self.client_cert)
        user_id_not_exist = 'user_id_not_exist'
        resp = self._get_access_token(
            client_id=user_id_not_exist,
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: '
            f'the user does not exist. user id: {user_id_not_exist}',
            self.log_fix.output,
        )

    def test_get_access_token_cert_dn_not_match_user_id(self):
        self._create_mapping()
        user, user_domain, _ = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id') + "_diff",
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: {} check failed. '
            'DN value: {}, DB value: {}.'.format(
                'user id', user.get('id') + '_diff', user.get('id')
            ),
            self.log_fix.output,
        )

    def test_get_access_token_cert_dn_not_match_user_name(self):
        self._create_mapping()
        user, user_domain, _ = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name') + "_diff",
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: {} check failed. '
            'DN value: {}, DB value: {}.'.format(
                'user name', user.get('name') + '_diff', user.get('name')
            ),
            self.log_fix.output,
        )

    def test_get_access_token_cert_dn_not_match_email(self):
        self._create_mapping()
        user, user_domain, _ = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                email_address=user.get('email') + "_diff",
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: {} check failed. '
            'DN value: {}, DB value: {}.'.format(
                'user email', user.get('email') + '_diff', user.get('email')
            ),
            self.log_fix.output,
        )

    def test_get_access_token_cert_dn_not_match_domain_id(self):
        self._create_mapping()
        user, user_domain, _ = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id') + "_diff",
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: {} check failed. '
            'DN value: {}, DB value: {}.'.format(
                'user domain id',
                user_domain.get('id') + '_diff',
                user_domain.get('id'),
            ),
            self.log_fix.output,
        )

    def test_get_access_token_cert_dn_not_match_domain_name(self):
        self._create_mapping()
        user, user_domain, _ = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name') + "_diff",
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: {} check failed. '
            'DN value: {}, DB value: {}.'.format(
                'user domain name',
                user_domain.get('name') + '_diff',
                user_domain.get('name'),
            ),
            self.log_fix.output,
        )

    def test_get_access_token_cert_dn_missing_user_id(self):
        self._create_mapping()
        user, user_domain, _ = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: mapping rule process failed.',
            self.log_fix.output,
        )

    def test_get_access_token_cert_dn_missing_user_name(self):
        self._create_mapping()
        user, user_domain, _ = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: mapping rule process failed.',
            self.log_fix.output,
        )

    def test_get_access_token_cert_dn_missing_email(self):
        self._create_mapping()
        user, user_domain, _ = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                domain_component=user_domain.get('id'),
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: mapping rule process failed.',
            self.log_fix.output,
        )

    def test_get_access_token_cert_dn_missing_domain_id(self):
        self._create_mapping()
        user, user_domain, _ = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                email_address=user.get('email'),
                organization_name=user_domain.get('name'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: mapping rule process failed.',
            self.log_fix.output,
        )

    def test_get_access_token_cert_dn_missing_domain_name(self):
        self._create_mapping()
        user, user_domain, _ = self._create_project_user()
        *_, client_cert, _ = self._create_certificates(
            client_dn=unit.create_dn(
                country_name='jp',
                state_or_province_name='kanagawa',
                locality_name='kawasaki',
                organizational_unit_name='test',
                user_id=user.get('id'),
                common_name=user.get('name'),
                email_address=user.get('email'),
                domain_component=user_domain.get('id'),
            )
        )

        cert_content = self._get_cert_content(client_cert)
        resp = self._get_access_token(
            client_id=user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        self.assertUnauthorizedResp(resp)
        self.assertIn(
            'Get OAuth2.0 Access Token API: mapping rule process failed.',
            self.log_fix.output,
        )

    @mock.patch.object(Manager, 'issue_token')
    def test_get_access_token_issue_token_ks_error_400(self, mock_issue_token):
        self._create_mapping()
        err_msg = 'Boom!'
        mock_issue_token.side_effect = exception.ValidationError(err_msg)
        cert_content = self._get_cert_content(self.client_cert)
        resp = self._get_access_token(
            client_id=self.oauth2_user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.BAD_REQUEST,
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertEqual('invalid_request', json_resp['error'])
        self.assertEqual(err_msg, json_resp['error_description'])
        self.assertIn(err_msg, self.log_fix.output)

    @mock.patch.object(Manager, 'issue_token')
    def test_get_access_token_issue_token_ks_error_401(self, mock_issue_token):
        self._create_mapping()
        err_msg = 'Boom!'
        mock_issue_token.side_effect = exception.Unauthorized(err_msg)
        cert_content = self._get_cert_content(self.client_cert)
        resp = self._get_access_token(
            client_id=self.oauth2_user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.UNAUTHORIZED,
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertEqual('invalid_client', json_resp['error'])
        self.assertEqual(
            'The request you have made requires authentication.',
            json_resp['error_description'],
        )

    @mock.patch.object(Manager, 'issue_token')
    def test_get_access_token_issue_token_ks_error_other(
        self, mock_issue_token
    ):
        self._create_mapping()
        err_msg = 'Boom!'
        mock_issue_token.side_effect = exception.NotImplemented(err_msg)
        cert_content = self._get_cert_content(self.client_cert)
        resp = self._get_access_token(
            client_id=self.oauth2_user.get('id'),
            client_cert_content=cert_content,
            expected_status=exception.NotImplemented.code,
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertEqual('other_error', json_resp['error'])
        self.assertEqual(
            'An unknown error occurred and failed to get an OAuth2.0 '
            'access token.',
            json_resp['error_description'],
        )

    @mock.patch.object(Manager, 'issue_token')
    def test_get_access_token_issue_token_other_exception(
        self, mock_issue_token
    ):
        self._create_mapping()
        err_msg = 'Boom!'
        mock_issue_token.side_effect = Exception(err_msg)
        cert_content = self._get_cert_content(self.client_cert)
        resp = self._get_access_token(
            client_id=self.oauth2_user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.INTERNAL_SERVER_ERROR,
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertEqual('other_error', json_resp['error'])
        self.assertEqual(err_msg, json_resp['error_description'])

    @mock.patch.object(RuleProcessor, 'process')
    def test_get_access_token_process_other_exception(self, mock_process):
        self._create_mapping()
        err_msg = 'Boom!'
        mock_process.side_effect = Exception(err_msg)
        cert_content = self._get_cert_content(self.client_cert)
        resp = self._get_access_token(
            client_id=self.oauth2_user.get('id'),
            client_cert_content=cert_content,
            expected_status=http.client.INTERNAL_SERVER_ERROR,
        )
        LOG.debug(resp)
        json_resp = jsonutils.loads(resp.body)
        self.assertEqual('other_error', json_resp['error'])
        self.assertEqual(err_msg, json_resp['error_description'])
        self.assertIn(
            'Get OAuth2.0 Access Token API: mapping rule process failed.',
            self.log_fix.output,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils

default_domain_id = cfg.StrOpt(
    'default_domain_id',
    default='default',
    help=utils.fmt(
        """
This references the domain to use for all Identity API v2 requests (which are
not aware of domains). A domain with this ID can optionally be created for you
by `keystone-manage bootstrap`. The domain referenced by this ID cannot be
deleted on the v3 API, to prevent accidentally breaking the v2 API. There is
nothing special about this domain, other than the fact that it must exist to
order to maintain support for your v2 clients. There is typically no reason to
change this value.
"""
    ),
)

domain_specific_drivers_enabled = cfg.BoolOpt(
    'domain_specific_drivers_enabled',
    default=False,
    help=utils.fmt(
        """
A subset (or all) of domains can have their own identity driver, each with
their own partial configuration options, stored in either the resource backend
or in a file in a domain configuration directory (depending on the setting of
`[identity] domain_configurations_from_database`). Only values specific to the
domain need to be specified in this manner. This feature is disabled by
default, but may be enabled by default in a future release; set to true to
enable.
"""
    ),
)

domain_configurations_from_database = cfg.BoolOpt(
    'domain_configurations_from_database',
    default=False,
    help=utils.fmt(
        """
By default, domain-specific configuration data is read from files in the
directory identified by `[identity] domain_config_dir`. Enabling this
configuration option allows you to instead manage domain-specific
configurations through the API, which are then persisted in the backend
(typically, a SQL database), rather than using configuration files on disk.
"""
    ),
)

domain_config_dir = cfg.StrOpt(
    'domain_config_dir',
    default='/etc/keystone/domains',
    help=utils.fmt(
        """
Absolute path where keystone should locate domain-specific `[identity]`
configuration files. This option has no effect unless `[identity]
domain_specific_drivers_enabled` is set to true. There is typically no reason
to change this value.
"""
    ),
)

driver = cfg.StrOpt(
    'driver',
    default='sql',
    help=utils.fmt(
        """
Entry point for the identity backend driver in the `keystone.identity`
namespace. Keystone provides a `sql` and `ldap` driver. This option is also
used as the default driver selection (along with the other configuration
variables in this section) in the event that `[identity]
domain_specific_drivers_enabled` is enabled, but no applicable domain-specific
configuration is defined for the domain in question. Unless your deployment
primarily relies on `ldap` AND is not using domain-specific configuration, you
should typically leave this set to `sql`.
"""
    ),
)

caching = cfg.BoolOpt(
    'caching',
    default=True,
    help=utils.fmt(
        """
Toggle for identity caching. This has no effect unless global caching is
enabled. There is typically no reason to disable this.
"""
    ),
)

cache_time = cfg.IntOpt(
    'cache_time',
    default=600,
    help=utils.fmt(
        """
Time to cache identity data (in seconds). This has no effect unless global and
identity caching are enabled.
"""
    ),
)

max_password_length = cfg.IntOpt(
    'max_password_length',
    default=4096,
    max=4096,
    help=utils.fmt(
        """
Maximum allowed length for user passwords. Decrease this value to improve
performance. Changing this value does not effect existing passwords. This value
can also be overridden by certain hashing algorithms maximum allowed length
which takes precedence over the configured value.

The bcrypt max_password_length is 72 bytes.
"""
    ),
)

list_limit = cfg.IntOpt(
    'list_limit',
    help=utils.fmt(
        """
Maximum number of entities that will be returned in an identity collection.
"""
    ),
)

password_hash_algorithm = cfg.StrOpt(
    'password_hash_algorithm',
    choices=['bcrypt', 'bcrypt_sha256', 'scrypt', 'pbkdf2_sha512'],
    default='bcrypt',
    help=utils.fmt(
        """
The password hashing algorithm to use for passwords stored within keystone.
"""
    ),
)

password_hash_rounds = cfg.IntOpt(
    'password_hash_rounds',
    help=utils.fmt(
        """
This option represents a trade off between security and performance. Higher
values lead to slower performance, but higher security. Changing this option
will only affect newly created passwords as existing password hashes already
have a fixed number of rounds applied, so it is safe to tune this option in a
running cluster.

The default for bcrypt is 12, must be between 4 and 31, inclusive.

The default for scrypt is 16, must be within `range(1,32)`.

The default for pbkdf_sha512 is 60000, must be within `range(1,1<<32)`

WARNING: If using scrypt, increasing this value increases BOTH time AND
memory requirements to hash a password.
"""
    ),
)

salt_bytesize = cfg.IntOpt(
    'salt_bytesize',
    min=0,
    max=96,
    help=utils.fmt(
        """
Number of bytes to use in scrypt and pbkfd2_sha512 hashing salt.

Default for scrypt is 16 bytes.
Default for pbkfd2_sha512 is 16 bytes.

Limited to a maximum of 96 bytes due to the size of the column used to store
password hashes.
"""
    ),
)

scrypt_block_size = cfg.IntOpt(
    'scrypt_block_size',
    help=utils.fmt(
        """
Optional block size to pass to scrypt hash function (the `r` parameter).
Useful for tuning scrypt to optimal performance for your CPU architecture.
This option is only used when the `password_hash_algorithm` option is set
to `scrypt`. Defaults to 8.
"""
    ),
)

scrypt_paralellism = cfg.IntOpt(
    'scrypt_parallelism',
    help=utils.fmt(
        """
Optional parallelism to pass to scrypt hash function (the `p` parameter).
This option is only used when the `password_hash_algorithm` option is set
to `scrypt`. Defaults to 1.
"""
    ),
)

GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [
    default_domain_id,
    domain_specific_drivers_enabled,
    domain_configurations_from_database,
    domain_config_dir,
    driver,
    caching,
    cache_time,
    max_password_length,
    list_limit,
    password_hash_algorithm,
    password_hash_rounds,
    scrypt_block_size,
    scrypt_paralellism,
    salt_bytesize,
]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import functools
import http.client
import re
import wsgiref.util

from keystonemiddleware import auth_token
import oslo_i18n
from oslo_log import log
from oslo_serialization import jsonutils
import webob.dec
import webob.exc

from keystone.common import authorization
from keystone.common import context
from keystone.common import provider_api
from keystone.common import render_token
from keystone.common import tokenless_auth
from keystone.common import utils
import keystone.conf
from keystone import exception
from keystone.federation import constants as federation_constants
from keystone.federation import utils as federation_utils
from keystone.i18n import _
from keystone.models import token_model

CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)
PROVIDERS = provider_api.ProviderAPIs

# Environment variable used to pass the request context
CONTEXT_ENV = 'openstack.context'

__all__ = ('AuthContextMiddleware',)


CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)


JSON_ENCODE_CONTENT_TYPES = {'application/json', 'application/json-home'}

# minimum access rules support
ACCESS_RULES_MIN_VERSION = token_model.ACCESS_RULES_MIN_VERSION


def best_match_language(req):
    """Determine the best available locale.

    This returns best available locale based on the Accept-Language HTTP
    header passed in the request.
    """
    if not req.accept_language:
        return None
    return req.accept_language.best_match(
        oslo_i18n.get_available_languages('keystone')
    )


def base_url(context):
    url = CONF['public_endpoint']

    if not url:
        if 'environment' not in context:
            raise ValueError('Endpoint cannot be detected')
        url = wsgiref.util.application_uri(context['environment'])
        # remove version from the URL as it may be part of SCRIPT_NAME but
        # it should not be part of base URL
        url = re.sub(r'/v(3|(2\.0))/*$', '', url)

        # now remove the standard port
        url = utils.remove_standard_port(url)

    return url.rstrip('/')


def middleware_exceptions(method):
    @functools.wraps(method)
    def _inner(self, request):
        try:
            return method(self, request)
        except exception.Error as e:
            LOG.warning(e)
            return render_exception(
                e, request=request, user_locale=best_match_language(request)
            )
        except TypeError as e:
            LOG.exception(e)
            return render_exception(
                exception.ValidationError(e),
                request=request,
                user_locale=best_match_language(request),
            )
        except Exception as e:
            LOG.exception(e)
            return render_exception(
                exception.UnexpectedError(exception=e),
                request=request,
                user_locale=best_match_language(request),
            )

    return _inner


def render_response(body=None, status=None, headers=None, method=None):
    """Form a WSGI response."""
    if headers is None:
        headers = []
    else:
        headers = list(headers)
    headers.append(('Vary', 'X-Auth-Token'))

    if body is None:
        body = b''
        status = status or (
            http.client.NO_CONTENT,
            http.client.responses[http.client.NO_CONTENT],
        )
    else:
        content_types = [v for h, v in headers if h == 'Content-Type']
        if content_types:
            content_type = content_types[0]
        else:
            content_type = None

        if content_type is None or content_type in JSON_ENCODE_CONTENT_TYPES:
            body = jsonutils.dump_as_bytes(body, cls=utils.SmarterEncoder)
            if content_type is None:
                headers.append(('Content-Type', 'application/json'))
        status = status or (
            http.client.OK,
            http.client.responses[http.client.OK],
        )

    # NOTE(davechen): `mod_wsgi` follows the standards from pep-3333 and
    # requires the value in response header to be binary type(str) on python2,
    # unicode based string(str) on python3, or else keystone will not work
    # under apache with `mod_wsgi`.
    # keystone needs to check the data type of each header and convert the
    # type if needed.
    # see bug:
    # https://bugs.launchpad.net/keystone/+bug/1528981
    # see pep-3333:
    # https://www.python.org/dev/peps/pep-3333/#a-note-on-string-types
    # see source from mod_wsgi:
    # https://github.com/GrahamDumpleton/mod_wsgi(methods:
    # wsgi_convert_headers_to_bytes(...), wsgi_convert_string_to_bytes(...)
    # and wsgi_validate_header_value(...)).
    def _convert_to_str(headers):
        str_headers = []
        for header in headers:
            str_header = []
            for value in header:
                if not isinstance(value, str):
                    str_header.append(str(value))
                else:
                    str_header.append(value)
            # convert the list to the immutable tuple to build the headers.
            # header's key/value will be guaranteed to be str type.
            str_headers.append(tuple(str_header))
        return str_headers

    headers = _convert_to_str(headers)

    resp = webob.Response(
        body=body, status='%d %s' % status, headerlist=headers, charset='utf-8'
    )

    if method and method.upper() == 'HEAD':
        # NOTE(morganfainberg): HEAD requests should return the same status
        # as a GET request and same headers (including content-type and
        # content-length). The webob.Response object automatically changes
        # content-length (and other headers) if the body is set to b''. Capture
        # all headers and reset them on the response object after clearing the
        # body. The body can only be set to a binary-type (not TextType or
        # NoneType), so b'' is used here and should be compatible with
        # both py2x and py3x.
        stored_headers = resp.headers.copy()
        resp.body = b''
        for header, value in stored_headers.items():
            resp.headers[header] = value

    return resp


def render_exception(error, context=None, request=None, user_locale=None):
    """Form a WSGI response based on the current error."""
    error_message = error.args[0]
    message = oslo_i18n.translate(error_message, desired_locale=user_locale)
    if message is error_message:
        # translate() didn't do anything because it wasn't a Message,
        # convert to a string.
        message = str(message)

    body = {
        'error': {'code': error.code, 'title': error.title, 'message': message}
    }
    headers = []
    if isinstance(error, exception.AuthPluginException):
        body['error']['identity'] = error.authentication
    elif isinstance(error, exception.Unauthorized):
        # NOTE(gyee): we only care about the request environment in the
        # context. Also, its OK to pass the environment as it is read-only in
        # base_url()
        local_context = {}
        if request:
            local_context = {'environment': request.environ}
        elif context and 'environment' in context:
            local_context = {'environment': context['environment']}
        url = base_url(local_context)

        headers.append(('WWW-Authenticate', f'Keystone uri="{url}"'))
    return render_response(
        status=(error.code, error.title), body=body, headers=headers
    )


class AuthContextMiddleware(
    provider_api.ProviderAPIMixin, auth_token.BaseAuthProtocol
):
    """Build the authentication context from the request auth token."""

    kwargs_to_fetch_token = True

    def __init__(self, app):
        super().__init__(app, log=LOG, service_type='identity')
        self.token = None

    def fetch_token(self, token, **kwargs):
        try:
            self.token = self.token_provider_api.validate_token(
                token, access_rules_support=ACCESS_RULES_MIN_VERSION
            )
            return render_token.render_token_response_from_model(self.token)
        except exception.TokenNotFound:
            raise auth_token.InvalidToken(_('Could not find token'))

    def _build_tokenless_auth_context(self, request):
        """Build the authentication context.

        The context is built from the attributes provided in the env,
        such as certificate and scope attributes.
        """
        tokenless_helper = tokenless_auth.TokenlessAuthHelper(request.environ)

        (domain_id, project_id, trust_ref, unscoped, system) = (
            tokenless_helper.get_scope()
        )
        user_ref = tokenless_helper.get_mapped_user(project_id, domain_id)

        # NOTE(gyee): if it is an ephemeral user, the
        # given X.509 SSL client cert does not need to map to
        # an existing user.
        if user_ref['type'] == federation_utils.UserType.EPHEMERAL:
            auth_context = {}
            auth_context['group_ids'] = user_ref['group_ids']
            auth_context[federation_constants.IDENTITY_PROVIDER] = user_ref[
                federation_constants.IDENTITY_PROVIDER
            ]
            auth_context[federation_constants.PROTOCOL] = user_ref[
                federation_constants.PROTOCOL
            ]
            if domain_id and project_id:
                msg = _('Scoping to both domain and project is not allowed')
                raise ValueError(msg)
            if domain_id:
                auth_context['domain_id'] = domain_id
            if project_id:
                auth_context['project_id'] = project_id
            auth_context['roles'] = user_ref['roles']
        else:
            # it's the local user, so token data is needed.
            token = token_model.TokenModel()
            token.user_id = user_ref['id']
            token.methods = [CONF.tokenless_auth.protocol]
            token.domain_id = domain_id
            token.project_id = project_id

            auth_context = {'user_id': user_ref['id']}
            auth_context['is_delegated_auth'] = False
            if domain_id:
                auth_context['domain_id'] = domain_id
            if project_id:
                auth_context['project_id'] = project_id
            auth_context['roles'] = [role['name'] for role in token.roles]
        return auth_context

    def _validate_trusted_issuer(self, request):
        """To further filter the certificates that are trusted.

        If the config option 'trusted_issuer' is absent or does
        not contain the trusted issuer DN, no certificates
        will be allowed in tokenless authorization.

        :param env: The env contains the client issuer's attributes
        :type env: dict
        :returns: True if client_issuer is trusted; otherwise False
        """
        if not CONF.tokenless_auth.trusted_issuer:
            return False

        issuer = request.environ.get(CONF.tokenless_auth.issuer_attribute)
        if not issuer:
            msg = (
                'Cannot find client issuer in env by the '
                'issuer attribute - %s.'
            )
            LOG.info(msg, CONF.tokenless_auth.issuer_attribute)
            return False

        if issuer in CONF.tokenless_auth.trusted_issuer:
            return True

        msg = (
            'The client issuer %(client_issuer)s does not match with '
            'the trusted issuer %(trusted_issuer)s'
        )
        LOG.info(
            msg,
            {
                'client_issuer': issuer,
                'trusted_issuer': CONF.tokenless_auth.trusted_issuer,
            },
        )

        return False

    @middleware_exceptions
    def process_request(self, request):
        context_env = request.environ.get(CONTEXT_ENV, {})

        # NOTE(notmorgan): This code is merged over from the admin token
        # middleware and now emits the security warning when the
        # conf.admin_token value is set.
        token = request.headers.get(authorization.AUTH_TOKEN_HEADER)
        if CONF.admin_token and (token == CONF.admin_token):
            context_env['is_admin'] = True
            LOG.warning(
                "The use of the '[DEFAULT] admin_token' configuration"
                "option presents a significant security risk and should "
                "not be set. This option is deprecated in favor of using "
                "'keystone-manage bootstrap' and will be removed in a "
                "future release."
            )
            request.environ[CONTEXT_ENV] = context_env

        if not context_env.get('is_admin', False):
            resp = super().process_request(request)
            if resp:
                return resp
            if (
                request.token_auth.has_user_token
                and not request.user_token_valid
            ):
                raise exception.Unauthorized(_('Not authorized.'))
            if request.token_auth.user is not None:
                request.set_user_headers(request.token_auth.user)

        # NOTE(jamielennox): function is split so testing can check errors from
        # fill_context. There is no actual reason for fill_context to raise
        # errors rather than return a resp, simply that this is what happened
        # before refactoring and it was easier to port. This can be fixed up
        # and the middleware_exceptions helper removed.
        self.fill_context(request)

    def _keystone_specific_values(self, token, request_context):
        request_context.token_reference = (
            render_token.render_token_response_from_model(token)
        )
        if token.domain_scoped:
            # Domain scoped tokens should never have is_admin_project set
            # Even if KSA defaults it otherwise.  The two mechanisms are
            # parallel; only one or the other should be used for access.
            request_context.is_admin_project = False
            request_context.domain_id = token.domain_id
            request_context.domain_name = token.domain['name']
        if token.oauth_scoped:
            request_context.is_delegated_auth = True
            request_context.oauth_consumer_id = token.access_token[
                'consumer_id'
            ]
            request_context.oauth_access_token_id = token.access_token_id
        if token.trust_scoped:
            request_context.is_delegated_auth = True
            request_context.trust_id = token.trust_id
        if token.is_federated:
            request_context.group_ids = []
            for group in token.federated_groups:
                request_context.group_ids.append(group['id'])
        else:
            request_context.group_ids = []

    def fill_context(self, request):
        # The request context stores itself in thread-local memory for logging.

        if authorization.AUTH_CONTEXT_ENV in request.environ:
            msg = (
                'Auth context already exists in the request '
                'environment; it will be used for authorization '
                'instead of creating a new one.'
            )
            LOG.warning(msg)
            return

        kwargs = {'authenticated': False, 'overwrite': True}
        request_context = context.RequestContext.from_environ(
            request.environ, **kwargs
        )
        request.environ[context.REQUEST_CONTEXT_ENV] = request_context

        # NOTE(gyee): token takes precedence over SSL client certificates.
        # This will preserve backward compatibility with the existing
        # behavior. Tokenless authorization with X.509 SSL client
        # certificate is effectively disabled if no trusted issuers are
        # provided.

        if request.environ.get(CONTEXT_ENV, {}).get('is_admin', False):
            request_context.is_admin = True
            auth_context = {}

        elif request.token_auth.has_user_token:
            # Keystone enforces policy on some values that other services
            # do not, and should not, use.  This adds them in to the context.
            if not self.token:
                self.token = PROVIDERS.token_provider_api.validate_token(
                    request.user_token,
                    access_rules_support=request.headers.get(
                        authorization.ACCESS_RULES_HEADER
                    ),
                )
            self._keystone_specific_values(self.token, request_context)
            request_context.auth_token = request.user_token
            auth_context = request_context.to_policy_values()
            additional = {
                'trust_id': request_context.trust_id,
                'trustor_id': request_context.trustor_id,
                'trustee_id': request_context.trustee_id,
                'domain_id': request_context.domain_id,
                'domain_name': request_context.domain_name,
                'group_ids': request_context.group_ids,
                'token': self.token,
            }
            auth_context.update(additional)

        elif self._validate_trusted_issuer(request):
            auth_context = self._build_tokenless_auth_context(request)
            # NOTE(gyee): we are no longer using auth_context when formulating
            # the credentials for RBAC. Instead, we are using the (Oslo)
            # request context. So we'll need to set all the necessary
            # credential attributes in the request context here.
            token_attributes = frozenset(
                (
                    'user_id',
                    'project_id',
                    'domain_id',
                    'user_domain_id',
                    'project_domain_id',
                    'user_domain_name',
                    'project_domain_name',
                    'roles',
                    'is_admin',
                    'project_name',
                    'domain_name',
                    'system_scope',
                    'is_admin_project',
                    'service_user_id',
                    'service_user_name',
                    'service_project_id',
                    'service_project_name',
                    'service_user_domain_id',
                    'service_user_domain_name',
                    'service_project_domain_id',
                    'service_project_domain_name',
                    'service_roles',
                )
            )
            for attr in token_attributes:
                if attr in auth_context:
                    setattr(request_context, attr, auth_context[attr])
            # NOTE(gyee): request_context.token_reference is always
            # expecting a 'token' key regardless. But in the case of X.509
            # tokenless auth, we don't need a token. So setting it to None
            # should be suffice.
            request_context.token_reference = {'token': None}
        else:
            # There is either no auth token in the request or the certificate
            # issuer is not trusted. No auth context will be set. This
            # typically happens on an initial token request.
            return

        # set authenticated to flag to keystone that a token has been validated
        request_context.authenticated = True

        LOG.debug('RBAC: auth_context: %s', auth_context)
        request.environ[authorization.AUTH_CONTEXT_ENV] = auth_context

    @classmethod
    def factory(cls, global_config, **local_config):
        """Used for loading in middleware (holdover from paste.deploy)."""

        def _factory(app):
            conf = global_config.copy()
            conf.update(local_config)
            return cls(app, **local_config)

        return _factory



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils

trusted_issuer = cfg.MultiStrOpt(
    'trusted_issuer',
    default=[],
    help=utils.fmt(
        """
The list of distinguished names which identify trusted issuers of client
certificates allowed to use X.509 tokenless authorization. If the option is
absent then no certificates will be allowed. The format for the values of a
distinguished name (DN) must be separated by a comma and contain no spaces.
Furthermore, because an individual DN may contain commas, this configuration
option may be repeated multiple times to represent multiple values. For
example, keystone.conf would include two consecutive lines in order to trust
two different DNs, such as `trusted_issuer = CN=john,OU=keystone,O=openstack`
and `trusted_issuer = CN=mary,OU=eng,O=abc`.
"""
    ),
)

protocol = cfg.StrOpt(
    'protocol',
    default='x509',
    help=utils.fmt(
        """
The federated protocol ID used to represent X.509 tokenless authorization. This
is used in combination with the value of `[tokenless_auth] issuer_attribute` to
find a corresponding federated mapping. In a typical deployment, there is no
reason to change this value.
"""
    ),
)

issuer_attribute = cfg.StrOpt(
    'issuer_attribute',
    default='SSL_CLIENT_I_DN',
    help=utils.fmt(
        """
The name of the WSGI environment variable used to pass the issuer of the client
certificate to keystone. This attribute is used as an identity provider ID
for the X.509 tokenless authorization along with the protocol to look up its
corresponding mapping. In a typical deployment, there is no reason to change
this value.
"""
    ),
)


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [trusted_issuer, protocol, issuer_attribute]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Copyright 2014 IBM Corp.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import sys
import warnings

from oslo_log import log
from sqlalchemy import exc
from testtools import matchers

from keystone.tests import unit

LOG = log.getLogger(__name__)


class BaseTestTestCase(unit.BaseTestCase):
    def test_unexpected_exit(self):
        # if a test calls sys.exit it raises rather than exiting.
        self.assertThat(
            lambda: sys.exit(), matchers.raises(unit.UnexpectedExit)
        )


class TestOverrideSkipping(unit.BaseTestCase):
    class TestParent(unit.BaseTestCase):
        def test_in_parent(self):
            pass

    class TestChild(TestParent):
        def test_in_parent(self):
            self.skip_test_overrides('some message')

        def test_not_in_parent(self):
            self.skip_test_overrides('some message')

    def test_skip_test_override_success(self):
        # NOTE(dstanek): let's run the test and see what happens
        test = self.TestChild('test_in_parent')
        result = test.run()

        # NOTE(dstanek): reach into testtools to ensure the test succeeded
        self.assertEqual([], result.decorated.errors)

    def test_skip_test_override_fails_for_missing_parent_test_case(self):
        # NOTE(dstanek): let's run the test and see what happens
        test = self.TestChild('test_not_in_parent')
        result = test.run()

        # NOTE(dstanek): reach into testtools to ensure the test failed
        #                the way we expected
        observed_error = result.decorated.errors[0]
        observed_error_msg = observed_error[1]
        expected_error_msg = (
            "'test_not_in_parent' is not a previously defined test method"
        )
        self.assertIn(expected_error_msg, observed_error_msg)


class TestTestCase(unit.TestCase):
    def test_bad_log(self):
        # If the arguments are invalid for the string in a log it raises an
        # exception during testing.
        self.assertThat(
            lambda: LOG.warning('String %(p1)s %(p2)s', {'p1': 'something'}),
            matchers.raises(KeyError),
        )

    def test_sa_warning(self):
        self.assertThat(
            lambda: warnings.warn('test sa warning error', exc.SAWarning),
            matchers.raises(exc.SAWarning),
        )

    def test_deprecation_warnings_are_raised_as_exceptions_in_tests(self):
        self.assertThat(
            lambda: warnings.warn('this is deprecated', DeprecationWarning),
            matchers.raises(DeprecationWarning),
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Fixtures for Federation Mapping."""

EMPLOYEE_GROUP_ID = "0cd5e9"
CONTRACTOR_GROUP_ID = "85a868"
TESTER_GROUP_ID = "123"
TESTER_GROUP_NAME = "tester"
DEVELOPER_GROUP_ID = "xyz"
DEVELOPER_GROUP_NAME = "Developer"
CONTRACTOR_GROUP_NAME = "Contractor"
DEVELOPER_GROUP_DOMAIN_NAME = "outsourcing"
DEVELOPER_GROUP_DOMAIN_ID = "5abc43"
FEDERATED_DOMAIN = "Federated"
LOCAL_DOMAIN = "Local"

# Mapping summary:
# LastName Smith & Not Contractor or SubContractor -> group 0cd5e9
# FirstName Jill & Contractor or SubContractor -> to group 85a868
MAPPING_SMALL = {
    "rules": [
        {
            "local": [
                {"group": {"id": EMPLOYEE_GROUP_ID}},
                {"user": {"name": "{0}"}},
            ],
            "remote": [
                {"type": "UserName"},
                {
                    "type": "orgPersonType",
                    "not_any_of": ["Contractor", "SubContractor"],
                },
                {"type": "LastName", "any_one_of": ["Bo"]},
            ],
        },
        {
            "local": [
                {"group": {"id": CONTRACTOR_GROUP_ID}},
                {"user": {"name": "{0}"}},
            ],
            "remote": [
                {"type": "UserName"},
                {
                    "type": "orgPersonType",
                    "any_one_of": ["Contractor", "SubContractor"],
                },
                {"type": "FirstName", "any_one_of": ["Jill"]},
            ],
        },
    ]
}

# Mapping summary:
# orgPersonType Admin or Big Cheese -> name {0} {1} email {2} and group 0cd5e9
# orgPersonType Customer -> user name {0} email {1}
# orgPersonType Test and email ^@example.com$ -> group 123 and xyz
MAPPING_LARGE = {
    "rules": [
        {
            "local": [
                {
                    "user": {"name": "{0} {1}", "email": "{2}"},
                    "group": {"id": EMPLOYEE_GROUP_ID},
                }
            ],
            "remote": [
                {"type": "FirstName"},
                {"type": "LastName"},
                {"type": "Email"},
                {
                    "type": "orgPersonType",
                    "any_one_of": ["Admin", "Big Cheese"],
                },
            ],
        },
        {
            "local": [{"user": {"name": "{0}", "email": "{1}"}}],
            "remote": [
                {"type": "UserName"},
                {"type": "Email"},
                {
                    "type": "orgPersonType",
                    "not_any_of": [
                        "Admin",
                        "Employee",
                        "Contractor",
                        "Tester",
                    ],
                },
            ],
        },
        {
            "local": [
                {"group": {"id": TESTER_GROUP_ID}},
                {"group": {"id": DEVELOPER_GROUP_ID}},
                {"user": {"name": "{0}"}},
            ],
            "remote": [
                {"type": "UserName"},
                {"type": "orgPersonType", "any_one_of": ["Tester"]},
                {
                    "type": "Email",
                    "any_one_of": [".*@example.com$"],
                    "regex": True,
                },
            ],
        },
    ]
}

MAPPING_BAD_REQ = {
    "rules": [
        {
            "local": [{"user": "name"}],
            "remote": [{"type": "UserName", "bad_requirement": ["Young"]}],
        }
    ]
}

MAPPING_BAD_VALUE = {
    "rules": [
        {
            "local": [{"user": "name"}],
            "remote": [{"type": "UserName", "any_one_of": "should_be_list"}],
        }
    ]
}

MAPPING_NO_RULES: dict[str, list[str]] = {'rules': []}

MAPPING_NO_REMOTE = {"rules": [{"local": [{"user": "name"}], "remote": []}]}

MAPPING_MISSING_LOCAL = {
    "rules": [
        {"remote": [{"type": "UserName", "any_one_of": "should_be_list"}]}
    ]
}

MAPPING_WRONG_TYPE = {
    "rules": [
        {"local": [{"user": "{1}"}], "remote": [{"not_type": "UserName"}]}
    ]
}

MAPPING_MISSING_TYPE = {
    "rules": [{"local": [{"user": "{1}"}], "remote": [{}]}]
}

MAPPING_EXTRA_REMOTE_PROPS_NOT_ANY_OF = {
    "rules": [
        {
            "local": [{"group": {"id": "0cd5e9"}}, {"user": {"name": "{0}"}}],
            "remote": [
                {"type": "UserName"},
                {
                    "type": "orgPersonType",
                    "not_any_of": ["SubContractor"],
                    "invalid_type": "xyz",
                },
            ],
        }
    ]
}

MAPPING_EXTRA_REMOTE_PROPS_ANY_ONE_OF = {
    "rules": [
        {
            "local": [{"group": {"id": "0cd5e9"}}, {"user": {"name": "{0}"}}],
            "remote": [
                {"type": "UserName"},
                {
                    "type": "orgPersonType",
                    "any_one_of": ["SubContractor"],
                    "invalid_type": "xyz",
                },
            ],
        }
    ]
}

MAPPING_EXTRA_REMOTE_PROPS_JUST_TYPE = {
    "rules": [
        {
            "local": [{"group": {"id": "0cd5e9"}}, {"user": {"name": "{0}"}}],
            "remote": [
                {"type": "UserName"},
                {"type": "orgPersonType", "invalid_type": "xyz"},
            ],
        }
    ]
}

MAPPING_EXTRA_RULES_PROPS = {
    "rules": [
        {
            "local": [{"group": {"id": "0cd5e9"}}, {"user": {"name": "{0}"}}],
            "invalid_type": {"id": "xyz"},
            "remote": [
                {"type": "UserName"},
                {"type": "orgPersonType", "not_any_of": ["SubContractor"]},
            ],
        }
    ]
}

MAPPING_TESTER_REGEX = {
    "rules": [
        {
            "local": [{"user": {"name": "{0}"}}],
            "remote": [{"type": "UserName"}],
        },
        {
            "local": [{"group": {"id": TESTER_GROUP_ID}}],
            "remote": [
                {
                    "type": "orgPersonType",
                    "any_one_of": [".*Tester*"],
                    "regex": True,
                }
            ],
        },
    ]
}


MAPPING_DIRECT_MAPPING_THROUGH_KEYWORD = {
    "rules": [
        {
            "local": [
                {"user": {"name": "{0}"}},
                {"group": {"id": TESTER_GROUP_ID}},
            ],
            "remote": [{"type": "UserName", "any_one_of": ["bwilliams"]}],
        }
    ]
}

MAPPING_DEVELOPER_REGEX = {
    "rules": [
        {
            "local": [
                {"user": {"name": "{0}"}, "group": {"id": DEVELOPER_GROUP_ID}}
            ],
            "remote": [
                {"type": "UserName"},
                {"type": "orgPersonType", "any_one_of": ["Developer"]},
                {
                    "type": "Email",
                    "not_any_of": [".*@example.org$"],
                    "regex": True,
                },
            ],
        }
    ]
}

MAPPING_GROUP_NAMES = {
    "rules": [
        {
            "local": [{"user": {"name": "{0}"}}],
            "remote": [{"type": "UserName"}],
        },
        {
            "local": [
                {
                    "group": {
                        "name": DEVELOPER_GROUP_NAME,
                        "domain": {"name": DEVELOPER_GROUP_DOMAIN_NAME},
                    }
                }
            ],
            "remote": [{"type": "orgPersonType", "any_one_of": ["Employee"]}],
        },
        {
            "local": [
                {
                    "group": {
                        "name": TESTER_GROUP_NAME,
                        "domain": {"id": DEVELOPER_GROUP_DOMAIN_ID},
                    }
                }
            ],
            "remote": [{"type": "orgPersonType", "any_one_of": ["BuildingX"]}],
        },
    ]
}

MAPPING_GROUP_NAME_WITHOUT_DOMAIN = {
    "rules": [
        {
            "local": [{"group": {"name": DEVELOPER_GROUP_NAME}}],
            "remote": [{"type": "orgPersonType", "any_one_of": ["Employee"]}],
        }
    ]
}

MAPPING_GROUP_ID_WITH_DOMAIN = {
    "rules": [
        {
            "local": [
                {
                    "group": {
                        "id": EMPLOYEE_GROUP_ID,
                        "domain": {"id": DEVELOPER_GROUP_DOMAIN_ID},
                    }
                }
            ],
            "remote": [{"type": "orgPersonType", "any_one_of": ["Employee"]}],
        }
    ]
}

MAPPING_BAD_GROUP = {
    "rules": [
        {
            "local": [{"group": {}}],
            "remote": [{"type": "orgPersonType", "any_one_of": ["Employee"]}],
        }
    ]
}

MAPPING_BAD_DOMAIN = {
    "rules": [
        {
            "local": [
                {
                    "group": {
                        "id": EMPLOYEE_GROUP_ID,
                        "domain": {
                            "id": DEVELOPER_GROUP_DOMAIN_ID,
                            "badkey": "badvalue",
                        },
                    }
                }
            ],
            "remote": [{"type": "orgPersonType", "any_one_of": ["Employee"]}],
        }
    ]
}

MAPPING_EPHEMERAL_USER = {
    "rules": [
        {
            "local": [
                {
                    "user": {
                        "name": "{0}",
                        "domain": {"id": FEDERATED_DOMAIN},
                        "type": "ephemeral",
                    }
                }
            ],
            "remote": [
                {"type": "UserName"},
                {"type": "UserName", "any_one_of": ["tbo"]},
            ],
        }
    ]
}

MAPPING_EPHEMERAL_USER_REMOTE_DOMAIN = {
    "rules": [
        {
            "local": [
                {
                    "user": {
                        "name": "{0}",
                        "domain": {"name": "{1}"},
                        "type": "ephemeral",
                    }
                }
            ],
            "remote": [
                {"type": "UserName"},
                {"type": "OIDC-openstack-user-domain"},
            ],
        }
    ]
}

MAPPING_GROUPS_WHITELIST = {
    "rules": [
        {
            "remote": [
                {
                    "type": "orgPersonType",
                    "whitelist": ["Developer", "Contractor"],
                },
                {"type": "UserName"},
            ],
            "local": [
                {"groups": "{0}", "domain": {"id": DEVELOPER_GROUP_DOMAIN_ID}},
                {"user": {"name": "{1}"}},
            ],
        }
    ]
}

MAPPING_EPHEMERAL_USER_LOCAL_DOMAIN = {
    "rules": [
        {
            "local": [
                {
                    "user": {
                        "name": "{0}",
                        "domain": {"id": LOCAL_DOMAIN},
                        "type": "ephemeral",
                    }
                }
            ],
            "remote": [
                {"type": "UserName"},
                {"type": "UserName", "any_one_of": ["jsmith"]},
            ],
        }
    ]
}

MAPPING_GROUPS_WHITELIST_MISSING_DOMAIN = {
    "rules": [
        {
            "remote": [
                {
                    "type": "orgPersonType",
                    "whitelist": ["Developer", "Contractor"],
                }
            ],
            "local": [{"groups": "{0}"}],
        }
    ]
}

MAPPING_LOCAL_USER_LOCAL_DOMAIN = {
    "rules": [
        {
            "local": [
                {
                    "user": {
                        "name": "{0}",
                        "domain": {"id": LOCAL_DOMAIN},
                        "type": "local",
                    }
                }
            ],
            "remote": [
                {"type": "UserName"},
                {"type": "UserName", "any_one_of": ["jsmith"]},
            ],
        }
    ]
}

MAPPING_GROUPS_BLACKLIST_MULTIPLES = {
    "rules": [
        {
            "remote": [
                {
                    "type": "orgPersonType",
                    "blacklist": ["Developer", "Manager"],
                },
                {"type": "Thing"},  # this could be variable length!
                {"type": "UserName"},
            ],
            "local": [
                {"groups": "{0}", "domain": {"id": DEVELOPER_GROUP_DOMAIN_ID}},
                {"user": {"name": "{2}"}},
            ],
        }
    ]
}
MAPPING_GROUPS_BLACKLIST = {
    "rules": [
        {
            "remote": [
                {
                    "type": "orgPersonType",
                    "blacklist": ["Developer", "Manager"],
                },
                {"type": "UserName"},
            ],
            "local": [
                {"groups": "{0}", "domain": {"id": DEVELOPER_GROUP_DOMAIN_ID}},
                {"user": {"name": "{1}"}},
            ],
        }
    ]
}

MAPPING_GROUPS_BLACKLIST_REGEX = {
    "rules": [
        {
            "remote": [
                {
                    "type": "orgPersonType",
                    "blacklist": [".*Employee$"],
                    "regex": True,
                }
            ],
            "local": [{"groups": "{0}", "domain": {"id": FEDERATED_DOMAIN}}],
        }
    ]
}

MAPPING_GROUPS_WHITELIST_REGEX = {
    "rules": [
        {
            "remote": [
                {
                    "type": "orgPersonType",
                    "whitelist": [".*Employee$"],
                    "regex": True,
                }
            ],
            "local": [{"groups": "{0}", "domain": {"id": FEDERATED_DOMAIN}}],
        }
    ]
}

# Exercise all possibilities of user identification. Values are hardcoded on
# purpose.
MAPPING_USER_IDS = {
    "rules": [
        {
            "local": [{"user": {"name": "{0}"}}],
            "remote": [
                {"type": "UserName"},
                {"type": "UserName", "any_one_of": ["jsmith"]},
            ],
        },
        {
            "local": [
                {
                    "user": {
                        "name": "{0}",
                        "id": "abc123@example.com",
                        "domain": {"id": "federated"},
                    }
                }
            ],
            "remote": [
                {"type": "UserName"},
                {"type": "UserName", "any_one_of": ["tbo"]},
            ],
        },
        {
            "local": [{"user": {"id": "{0}"}}],
            "remote": [
                {"type": "UserName"},
                {"type": "UserName", "any_one_of": ["bob"]},
            ],
        },
        {
            "local": [
                {
                    "user": {
                        "id": "abc123@example.com",
                        "name": "{0}",
                        "domain": {"id": "federated"},
                    }
                }
            ],
            "remote": [
                {"type": "UserName"},
                {"type": "UserName", "any_one_of": ["bwilliams"]},
            ],
        },
    ]
}

MAPPING_GROUPS_BLACKLIST_MISSING_DOMAIN = {
    "rules": [
        {
            "remote": [
                {
                    "type": "orgPersonType",
                    "blacklist": ["Developer", "Manager"],
                }
            ],
            "local": [{"groups": "{0}"}],
        }
    ]
}

MAPPING_GROUPS_WHITELIST_AND_BLACKLIST = {
    "rules": [
        {
            "remote": [
                {
                    "type": "orgPersonType",
                    "blacklist": ["Employee"],
                    "whitelist": ["Contractor"],
                }
            ],
            "local": [
                {"groups": "{0}", "domain": {"id": DEVELOPER_GROUP_DOMAIN_ID}}
            ],
        }
    ]
}

# Mapping used by tokenless test cases, it maps the user_name
# and domain_name.
MAPPING_WITH_USERNAME_AND_DOMAINNAME = {
    'rules': [
        {
            'local': [
                {
                    'user': {
                        'name': '{0}',
                        'domain': {'name': '{1}'},
                        'type': 'local',
                    }
                }
            ],
            'remote': [
                {'type': 'SSL_CLIENT_USER_NAME'},
                {'type': 'SSL_CLIENT_DOMAIN_NAME'},
            ],
        }
    ]
}

# Mapping used by tokenless test cases, it maps the user_id
# and domain_name.
MAPPING_WITH_USERID_AND_DOMAINNAME = {
    'rules': [
        {
            'local': [
                {
                    'user': {
                        'id': '{0}',
                        'domain': {'name': '{1}'},
                        'type': 'local',
                    }
                }
            ],
            'remote': [
                {'type': 'SSL_CLIENT_USER_ID'},
                {'type': 'SSL_CLIENT_DOMAIN_NAME'},
            ],
        }
    ]
}

# Mapping used by tokenless test cases, it maps the user_name
# and domain_id.
MAPPING_WITH_USERNAME_AND_DOMAINID = {
    'rules': [
        {
            'local': [
                {
                    'user': {
                        'name': '{0}',
                        'domain': {'id': '{1}'},
                        'type': 'local',
                    }
                }
            ],
            'remote': [
                {'type': 'SSL_CLIENT_USER_NAME'},
                {'type': 'SSL_CLIENT_DOMAIN_ID'},
            ],
        }
    ]
}

# Mapping used by tokenless test cases, it maps the user_id
# and domain_id.
MAPPING_WITH_USERID_AND_DOMAINID = {
    'rules': [
        {
            'local': [
                {
                    'user': {
                        'id': '{0}',
                        'domain': {'id': '{1}'},
                        'type': 'local',
                    }
                }
            ],
            'remote': [
                {'type': 'SSL_CLIENT_USER_ID'},
                {'type': 'SSL_CLIENT_DOMAIN_ID'},
            ],
        }
    ]
}

# Mapping used by tokenless test cases, it maps the domain_id only.
MAPPING_WITH_DOMAINID_ONLY = {
    'rules': [
        {
            'local': [{'user': {'domain': {'id': '{0}'}, 'type': 'local'}}],
            'remote': [{'type': 'SSL_CLIENT_DOMAIN_ID'}],
        }
    ]
}

MAPPING_GROUPS_IDS_WHITELIST = {
    "rules": [
        {
            "local": [
                {"user": {"name": "{0}"}},
                {"group_ids": "{1}"},
                {"group": {"id": "{2}"}},
            ],
            "remote": [
                {"type": "name"},
                {
                    "type": "group_ids",
                    "whitelist": ["abc123", "ghi789", "321cba"],
                },
                {"type": "group"},
            ],
        }
    ]
}

MAPPING_GROUPS_IDS_BLACKLIST = {
    "rules": [
        {
            "local": [
                {"user": {"name": "{0}"}},
                {"group_ids": "{1}"},
                {"group": {"id": "{2}"}},
            ],
            "remote": [
                {"type": "name"},
                {"type": "group_ids", "blacklist": ["def456"]},
                {"type": "group"},
            ],
        }
    ]
}

# Mapping used by tokenless test cases, it maps the domain_name only.
MAPPING_WITH_DOMAINNAME_ONLY = {
    'rules': [
        {
            'local': [{'user': {'domain': {'name': '{0}'}, 'type': 'local'}}],
            'remote': [{'type': 'SSL_CLIENT_DOMAIN_NAME'}],
        }
    ]
}

# Mapping used by tokenless test cases, it maps the user_name only.
MAPPING_WITH_USERNAME_ONLY = {
    'rules': [
        {
            'local': [{'user': {'name': '{0}', 'type': 'local'}}],
            'remote': [{'type': 'SSL_CLIENT_USER_NAME'}],
        }
    ]
}

# Mapping used by tokenless test cases, it maps the user_id only.
MAPPING_WITH_USERID_ONLY = {
    'rules': [
        {
            'local': [{'user': {'id': '{0}', 'type': 'local'}}],
            'remote': [{'type': 'SSL_CLIENT_USER_ID'}],
        }
    ]
}

MAPPING_FOR_EPHEMERAL_USER = {
    'rules': [
        {
            'local': [
                {
                    'user': {'name': '{0}', 'type': 'ephemeral'},
                    'group': {'id': 'dummy'},
                }
            ],
            'remote': [{'type': 'SSL_CLIENT_USER_NAME'}],
        }
    ]
}

MAPPING_FOR_EPHEMERAL_USER_AND_GROUP_DOMAIN_NAME = {
    'rules': [
        {
            'local': [
                {
                    'user': {'name': '{0}', 'type': 'ephemeral'},
                    'group': {'name': 'dummy', 'domain': {'name': 'dummy'}},
                }
            ],
            'remote': [{'type': 'SSL_CLIENT_USER_NAME'}],
        }
    ]
}

MAPPING_FOR_DEFAULT_EPHEMERAL_USER = {
    'rules': [
        {
            'local': [{'user': {'name': '{0}'}, 'group': {'id': 'dummy'}}],
            'remote': [{'type': 'SSL_CLIENT_USER_NAME'}],
        }
    ]
}

MAPPING_GROUPS_WHITELIST_PASS_THROUGH = {
    "rules": [
        {
            "remote": [{"type": "UserName"}],
            "local": [
                {
                    "user": {
                        "name": "{0}",
                        "domain": {"id": DEVELOPER_GROUP_DOMAIN_ID},
                    }
                }
            ],
        },
        {
            "remote": [{"type": "orgPersonType", "whitelist": ['Developer']}],
            "local": [
                {"groups": "{0}", "domain": {"id": DEVELOPER_GROUP_DOMAIN_ID}}
            ],
        },
    ]
}

MAPPING_BAD_LOCAL_SETUP = {
    "rules": [
        {
            "local": [
                {
                    "user": {"name": "{0}", "domain": {"id": "default"}},
                    "whatisthis": "local",
                }
            ],
            "remote": [{"type": "UserName"}],
        }
    ]
}

MAPPING_BAD_LOCAL_TYPE_USER_IN_ASSERTION = {
    "rules": [
        {
            "local": [{"user": {"name": "{0}", "groups": "{1}"}}],
            "remote": [
                {"type": "openstack_user"},
                {"type": "openstack_groups"},
                {"type": "openstack_roles", "any_one_of": ["Admin"]},
            ],
        }
    ]
}

MAPPING_GROUPS_WITH_EMAIL = {
    "rules": [
        {
            "remote": [
                {"type": "groups"},
                {"type": "userEmail"},
                {"type": "UserName"},
            ],
            "local": [
                {"groups": "{0}", "domain": {"id": DEVELOPER_GROUP_DOMAIN_ID}},
                {"user": {"name": "{2}", "email": "{1}"}},
            ],
        }
    ]
}


MAPPING_GROUPS_DOMAIN_OF_USER = {
    "rules": [
        {
            "local": [{"user": {"name": "{0}"}}, {"groups": "{1}"}],
            "remote": [
                {"type": "openstack_user"},
                {"type": "openstack_groups"},
            ],
        }
    ]
}

EMPLOYEE_ASSERTION = {
    'Email': 'tim@example.com',
    'UserName': 'tbo',
    'FirstName': 'Tim',
    'LastName': 'Bo',
    'orgPersonType': 'Employee;BuildingX',
}

EMPLOYEE_PARTTIME_ASSERTION = {
    'Email': 'tim@example.com',
    'UserName': 'tbo',
    'FirstName': 'Tim',
    'LastName': 'Bo',
    'orgPersonType': 'Employee;PartTimeEmployee;Manager',
}

EMPLOYEE_ASSERTION_MULTIPLE_GROUPS = {
    'Email': 'tim@example.com',
    'UserName': 'tbo',
    'FirstName': 'Tim',
    'LastName': 'Bo',
    'orgPersonType': 'Developer;Manager;Contractor',
    'Thing': 'yes!;maybe!;no!!',
}

EMPLOYEE_ASSERTION_PREFIXED = {
    'PREFIX_Email': 'tim@example.com',
    'PREFIX_UserName': 'tbo',
    'PREFIX_FirstName': 'Tim',
    'PREFIX_LastName': 'Bo',
    'PREFIX_orgPersonType': 'SuperEmployee;BuildingX',
}

CONTRACTOR_ASSERTION = {
    'Email': 'jill@example.com',
    'UserName': 'jsmith',
    'FirstName': 'Jill',
    'LastName': 'Smith',
    'orgPersonType': 'Contractor;Non-Dev',
}

ADMIN_ASSERTION = {
    'Email': 'bob@example.com',
    'UserName': 'bob',
    'FirstName': 'Bob',
    'LastName': 'Thompson',
    'orgPersonType': 'Admin;Chief',
}

CUSTOMER_ASSERTION = {
    'Email': 'beth@example.com',
    'UserName': 'bwilliams',
    'FirstName': 'Beth',
    'LastName': 'Williams',
    'orgPersonType': 'Customer',
}

ANOTHER_CUSTOMER_ASSERTION = {
    'Email': 'mark@example.com',
    'UserName': 'markcol',
    'FirstName': 'Mark',
    'LastName': 'Collins',
    'orgPersonType': 'Managers;CEO;CTO',
}

TESTER_ASSERTION = {
    'Email': 'testacct@example.com',
    'UserName': 'testacct',
    'FirstName': 'Test',
    'LastName': 'Account',
    'orgPersonType': 'MadeupGroup;Tester;GroupX',
}

ANOTHER_TESTER_ASSERTION = {
    'Email': 'testacct@example.com',
    'UserName': 'IamTester',
}

BAD_TESTER_ASSERTION = {
    'Email': 'eviltester@example.org',
    'UserName': 'Evil',
    'FirstName': 'Test',
    'LastName': 'Account',
    'orgPersonType': 'Tester',
}

BAD_DEVELOPER_ASSERTION = {
    'Email': 'evildeveloper@example.org',
    'UserName': 'Evil',
    'FirstName': 'Develop',
    'LastName': 'Account',
    'orgPersonType': 'Developer',
}

MALFORMED_TESTER_ASSERTION = {
    'Email': 'testacct@example.com',
    'UserName': 'testacct',
    'FirstName': 'Test',
    'LastName': 'Account',
    'orgPersonType': 'Tester',
    'object': object(),
    'dictionary': dict(zip('teststring', range(10))),
    'tuple': tuple(range(5)),
}

DEVELOPER_ASSERTION = {
    'Email': 'developacct@example.com',
    'UserName': 'developacct',
    'FirstName': 'Develop',
    'LastName': 'Account',
    'orgPersonType': 'Developer',
}

CONTRACTOR_MALFORMED_ASSERTION = {
    'UserName': 'user',
    'FirstName': object(),
    'orgPersonType': 'Contractor',
}

LOCAL_USER_ASSERTION = {'UserName': 'marek', 'UserType': 'random'}

ANOTHER_LOCAL_USER_ASSERTION = {
    'UserName': 'marek',
    'Position': 'DirectorGeneral',
}

USER_NO_GROUPS_ASSERTION = {
    'Email': 'nogroupsuser1@example.org',
    'UserName': 'nogroupsuser1',
    'orgPersonType': 'NoGroupsOrg',
}

UNMATCHED_GROUP_ASSERTION = {
    'REMOTE_USER': 'Any Momoose',
    'REMOTE_USER_GROUPS': 'EXISTS;NO_EXISTS',
}

GROUP_IDS_ASSERTION = {
    'name': 'opilotte',
    'group_ids': 'abc123;def456;ghi789',
    'group': 'klm012',
}

GROUP_IDS_ASSERTION_ONLY_ONE_GROUP = {
    'name': 'opilotte',
    'group_ids': '321cba',
    'group': '210mlk',
}

UNICODE_NAME_ASSERTION = {
    'PFX_Email': 'jon@example.com',
    'PFX_UserName': 'jonkare',
    'PFX_FirstName': 'Jon Kåre',
    'PFX_LastName': 'Hellån',
    'PFX_orgPersonType': 'Admin;Chief',
}

GROUPS_ASSERTION_ONLY_ONE_GROUP = {
    'userEmail': 'jill@example.com',
    'UserName': 'jsmith',
    'groups': 'ALL USERS',
}

GROUPS_ASSERTION_ONLY_ONE_NUMERICAL_GROUP = {
    'userEmail': 'jill@example.com',
    'UserName': 'jsmith',
    'groups': '1234',
}

GROUPS_DOMAIN_ASSERTION = {
    'openstack_user': 'bwilliams',
    'openstack_user_domain': 'default',
    'openstack_roles': 'Admin',
    'openstack_groups': 'JSON:{"name":"group1","domain":{"name":"xxx"}};'
    'JSON:{"name":"group2","domain":{"name":"yyy"}}',
}

USER_WITH_DOMAIN_ASSERTION = {
    'UserName': 'marek',
    'OIDC-openstack-user-domain': 'user_domain',
}

MAPPING_UNICODE = {
    "rules": [
        {
            "local": [
                {
                    "user": {"name": "{0} {1}", "email": "{2}"},
                    "group": {"id": EMPLOYEE_GROUP_ID},
                }
            ],
            "remote": [
                {"type": "PFX_FirstName"},
                {"type": "PFX_LastName"},
                {"type": "PFX_Email"},
                {
                    "type": "PFX_orgPersonType",
                    "any_one_of": ["Admin", "Big Cheese"],
                },
            ],
        }
    ]
}

MAPPING_PROJECTS = {
    "rules": [
        {
            "local": [
                {"user": {"name": "{0}"}},
                {
                    "projects": [
                        {
                            "name": "Production",
                            "roles": [{"name": "observer"}],
                        },
                        {"name": "Staging", "roles": [{"name": "member"}]},
                        {
                            "name": "Project for {0}",
                            "roles": [{"name": "admin"}],
                        },
                    ]
                },
            ],
            "remote": [
                {"type": "UserName"},
                {"type": "Email"},
                {"type": "orgPersonType", "any_one_of": ["Employee"]},
            ],
        }
    ]
}

MAPPING_PROJECTS_WITHOUT_ROLES = {
    "rules": [
        {
            "local": [
                {
                    "user": {"name": "{0}"},
                    "projects": [
                        {"name": "a"},
                        {"name": "b"},
                        {"name": "Project for {0}"},
                    ],
                }
            ],
            "remote": [{"type": "UserName"}],
        }
    ]
}

MAPPING_PROJECTS_WITHOUT_NAME = {
    "rules": [
        {
            "local": [
                {
                    "user": {"name": "{0}"},
                    "projects": [
                        {"roles": [{"name": "observer"}]},
                        {"name": "Staging", "roles": [{"name": "member"}]},
                        {
                            "name": "Project for {0}",
                            "roles": [{"name": "admin"}],
                        },
                    ],
                }
            ],
            "remote": [{"type": "UserName"}],
        }
    ]
}



import os
import requests

KEYCLOAK_USERNAME = os.environ.get('KEYCLOAK_USERNAME')
KEYCLOAK_PASSWORD = os.environ.get('KEYCLOAK_PASSWORD')
KEYCLOAK_URL = os.environ.get('KEYCLOAK_URL')
HOST_IP = os.environ.get('HOST_IP', 'localhost')


class KeycloakClient:
    def __init__(self):
        self.session = requests.session()

    @staticmethod
    def construct_url(realm, path):
        return f'{KEYCLOAK_URL}/admin/realms/{realm}/{path}'

    @staticmethod
    def token_endpoint(realm):
        return f'{KEYCLOAK_URL}/realms/{realm}/protocol/openid-connect/token'

    def _admin_auth(self, realm):
        params = {
            'grant_type': 'password',
            'client_id': 'admin-cli',
            'username': KEYCLOAK_USERNAME,
            'password': KEYCLOAK_PASSWORD,
            'scope': 'openid',
        }
        r = requests.post(self.token_endpoint(realm), data=params).json()
        headers = {
            'Authorization': f"Bearer {r['access_token']}",
            'Content-Type': 'application/json',
        }
        self.session.headers.update(headers)
        return r

    def create_client(self, realm, client_id, client_secret, redirect_uris):
        self._admin_auth(realm)
        data = {
            'clientId': client_id,
            'secret': client_secret,
            'redirectUris': redirect_uris,
            'implicitFlowEnabled': True,
            'directAccessGrantsEnabled': True,
        }
        return self.session.post(
            self.construct_url(realm, 'clients'), json=data
        )


def main():
    c = KeycloakClient()

    redirect_uris = [
        f'http://{HOST_IP}/identity/v3/auth/OS-FEDERATION/identity_providers/sso/protocols/openid/websso',
        f'http://{HOST_IP}/identity/v3/auth/OS-FEDERATION/websso/openid',
    ]

    c.create_client('master', 'devstack', 'nomoresecret', redirect_uris)


if __name__ == "__main__":
    main()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import versionutils
from oslo_policy import policy

from keystone.common.policies import base

DEPRECATED_REASON = (
    "The OAUTH1 consumer API is now aware of system scope and default roles."
)

deprecated_get_consumer = policy.DeprecatedRule(
    name=base.IDENTITY % 'get_consumer',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)
deprecated_list_consumers = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_consumers',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)
deprecated_create_consumer = policy.DeprecatedRule(
    name=base.IDENTITY % 'create_consumer',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)
deprecated_update_consumer = policy.DeprecatedRule(
    name=base.IDENTITY % 'update_consumer',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)
deprecated_delete_consumer = policy.DeprecatedRule(
    name=base.IDENTITY % 'delete_consumer',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)


consumer_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_consumer',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='Show OAUTH1 consumer details.',
        operations=[
            {'path': '/v3/OS-OAUTH1/consumers/{consumer_id}', 'method': 'GET'}
        ],
        deprecated_rule=deprecated_get_consumer,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_consumers',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='List OAUTH1 consumers.',
        operations=[{'path': '/v3/OS-OAUTH1/consumers', 'method': 'GET'}],
        deprecated_rule=deprecated_list_consumers,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_consumer',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Create OAUTH1 consumer.',
        operations=[{'path': '/v3/OS-OAUTH1/consumers', 'method': 'POST'}],
        deprecated_rule=deprecated_create_consumer,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'update_consumer',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Update OAUTH1 consumer.',
        operations=[
            {
                'path': '/v3/OS-OAUTH1/consumers/{consumer_id}',
                'method': 'PATCH',
            }
        ],
        deprecated_rule=deprecated_update_consumer,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'delete_consumer',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Delete OAUTH1 consumer.',
        operations=[
            {
                'path': '/v3/OS-OAUTH1/consumers/{consumer_id}',
                'method': 'DELETE',
            }
        ],
        deprecated_rule=deprecated_delete_consumer,
    ),
]


def list_rules():
    return consumer_policies
